import doctest
import logging
import os
import shutil
import unittest

import pygeonlp.api as api

logger = logging.getLogger(__name__)

"""
doc_string に書かれているサンプルコードを doctest で検証するテスト。

このモジュールのテストを行なうには、次のコマンドを実行してください。

python -m unittest -v pygeonlp.tests.test_doctest
"""


def get_testdir():
    return os.path.abspath(os.path.join(os.getcwd(), 'apitest'))


def setup(test):
    testdir = get_testdir()
    os.environ['GEONLP_DB_DIR'] = testdir
    os.makedirs(testdir, 0o755, exist_ok=True)
    api.init(db_dir=testdir)

    # デフォルト以外の辞書が登録されていたら
    # データベースを作り直す
    default_dics = [
        'geonlp:geoshape-city',
        'geonlp:geoshape-pref',
        'geonlp:ksj-station-N02',
    ]

    dict_manager = api.dict_manager.DictManager(db_dir=testdir)
    installed_dictionaries = [
        x.get_identifier() for x in dict_manager.getDictionaries()]
    for identifier in installed_dictionaries:
        if identifier not in default_dics:
            dict_manager.clearDatabase()
            break

    dict_manager.setupBasicDatabase()
    api.init(db_dir=testdir)


def load_tests(loader, tests, ignore):
    shutil.rmtree(get_testdir(), ignore_errors=True)

    from pygeonlp.api import devtool, temporal_filter
    modules = [
        api,
        api.service, api.metadata, api.dictionary, api.node,
        api.parser, api.linker, devtool,
        api.filter, temporal_filter,
    ]

    try:
        from pygeonlp.api import spatial_filter
        spatial_filter.SpatialFilter.get_geometry_from_geojson(
            {'type': 'Point', 'coordinates': [139.6917337, 35.6895014]})
        modules.append(spatial_filter)
    except api.filter.FilterError:
        # Since GDAL is not installed, spatial_filter test will be skipped.
        pass

    flags = doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS

    for module in modules:
        tests.addTest(
            doctest.DocTestSuite(module, setUp=setup, optionflags=flags))

    return tests


if __name__ == "__main__":
    # logging.basicConfig(
    #     level=logging.DEBUG,
    #     format="%(asctime)s:%(levelname)s:%(filename)s:%(lineno)d:%(message)s")
    unittest.main(failfast=True)



from collections.abc import Iterable
from logging import getLogger
import os
import re
from typing import Union

from pygeonlp import capi

from pygeonlp.api.metadata import Metadata

logger = getLogger(__name__)


class ServiceError(RuntimeError):
    """
    サービス実行時に例外が起こると、このクラスが発生します。
    """
    pass


class Service(object):
    """
    文の解析や地名語の検索機能を提供する低水準APIサービスクラスです。

    Attributes
    ----------
    _dict_cache : dict
        辞書ID（整数）をキー、辞書identifier（文字列）を値にもつ
        マッピングテーブル（キャッシュ用）。
    capi_ma : pygeonlp.capi オブジェクト
        C 実装の拡張形態素解析機能を利用するためのオブジェクト。
    db_dir: str
        このサービスが利用するデータベースディレクトリのパス。
    """

    def __init__(self,
                 db_dir: Union[str, bytes, os.PathLike, None] = None,
                 geoword_rules: dict = {}, **options):
        """
        このサービスが利用する辞書や、各種設定を初期化します。

        Parameters
        ----------
        db_dir : PathLike, optional
            利用するデータベースディレクトリのパス。
            省略した場合は ``api.get_db_dir()`` で決定します。
        geoword_rules: dict, optional
            地名語を解析するルールセットを指定します。
        options : dict, optional
            その他の解析オプションを指定します。

        Note
        ----
        地名語抽出ルールとして指定できる項目は以下の通りです。

        suffix : list of str
            地名接尾語を定義します。通常、一般名詞などの前に地名がくる場合、
            地名修飾語として解析されます。
            例：千代田区立 -> 千代田（名詞・固有名詞・地名語）＋区立（名詞・一般）

            地名接尾語として「立」を定義すると、
            千代田区（名詞・固有名詞・地名語）＋立（名詞・接尾・地名語）と
            解析されるようになります。
            地名接尾辞は "立,リツ,リツ" のように表層形、読み、発音を
            カンマで区切った文字列として定義してください。
            デフォルト値は ["前,マエ,マエ", "内,ナイ,ナイ", "立,リツ,リツ",
            "境,サカイ,サカイ", "東,ヒガシ,ヒガシ", "西,ニシ,ニシ",
            "南,ミナミ,ミナミ", "北,キタ,キタ"] です。

        excluded_word : list of str
            非地名語を定義します。
            地名解析辞書に登録されている語は地名語として解析されます。
            例：本部に行く -> 本部（名詞・固有名詞・地名語）＋に（助詞）……
            しかし特定の語を地名語として解析したくない場合は非地名語として
            定義してください。
            デフォルト値は ["本部","一部","月"] です。
            環境変数 GEONLP_EXCLUDED_WORD でデフォルト値を設定できます。


        その他の解析オプションは以下の通りです。

        address_class : str
            住所要素とみなす固有名クラスを正規表現で指定します。
            たとえば固有名クラスが「都道府県」である地名語から始まる住所表記だけを
            住所として解析したい（市区町村から始まる場合は無視したい）場合は
            r"^都道府県" を指定します。
            デフォルト値は r"^(都道府県|市区町村|行政地域|居住地名)(/.+|)" です。
            環境変数 GEONLP_ADRESS_CLASS でデフォルト値を設定できます。

        system_dic_dir : str
            MeCab システム辞書のディレクトリを指定します。
            省略した場合はデフォルトシステム辞書を利用します。
            環境変数 GEONLP_MECAB_DIC_DIR でデフォルト値を設定できます。
        """
        self._dict_cache = {}
        self.options = options
        self.capi_ma = None

        if db_dir is None:
            from . import get_db_dir
            db_dir = get_db_dir()

        self.db_dir = db_dir

        # データベースディレクトリの存在チェック
        if not os.path.exists(db_dir):
            raise RuntimeError(
                ("データベースディレクトリ {} がありません。".format(db_dir),
                 "setup_basic_database() で作成してください。"))

        # capi.ma 用のオプションを構築
        capi_options = {'data_dir': str(db_dir)}
        if 'suffix' in geoword_rules:
            suffix = geoword_rules["suffix"]
            if isinstance(suffix, str):
                capi_options['suffix'] = suffix
            elif isinstance(geoword_rules['suffix'], Iterable):
                capi_options['suffix'] = '|'.join(list(suffix))
            else:
                raise TypeError(
                    "'suffix' は文字列またはリストで指定してください。")

        excluded_word = geoword_rules.get(
            "excluded_word",
            os.environ.get("GEONLP_EXCLUDED_WORD", None)
        )
        if isinstance(excluded_word, str):
            capi_options['non_geoword'] = excluded_word
        elif isinstance(excluded_word, Iterable):
            capi_options['non_geoword'] = '|'.join(list(excluded_word))
        elif excluded_word is not None:
            raise TypeError(
                "'excluded_word' は文字列またはリストで指定してください。")

        # その他の解析オプション
        address_class = self.options.get(
            "address_class",
            os.environ.get(
                "GEONLP_ADDRESS_CLASS",
                r'^(都道府県|市区町村|行政地域|居住地名)(\/.+|)'
            )
        )
        if isinstance(address_class, str):
            if address_class == "":
                address_class = "^$"

            capi_options['address_regex'] = address_class
        else:
            raise TypeError(
                "'address_class' は正規表現文字列で指定してください。")

        system_dic_dir = self.options.get(
            "system_dic_dir",
            os.environ.get("GEONLP_MECAB_DIC_DIR", None)
        )
        if system_dic_dir is None:
            pass
        elif isinstance(system_dic_dir, str):
            capi_options['system_dic_dir'] = system_dic_dir
        else:
            raise TypeError(
                "'system_dic_dir' は文字列で指定してください。")

        self.capi_ma = capi.MA(capi_options)

    def ma_parse(self, sentence):
        """
        センテンスを形態素解析した結果を MeCab 互換の文字列として返します。

        Parameters
        ----------
        sentence : str
            解析する文字列。

        Returns
        -------
        str
            解析結果の改行区切りテキスト。

        Examples
        --------
        >>> from pygeonlp.api.service import Service
        >>> service = Service()
        >>> print(service.ma_parse('今日は国会議事堂前まで歩きました。'))
        <BLANKLINE>
        今日    名詞,副詞可能,*,*,*,*,今日,キョウ,キョー
        は      助詞,係助詞,*,*,*,*,は,ハ,ワ
        国会議事堂前    名詞,固有名詞,地名語,Bn4q6d:国会議事堂前駅/cE8W4w:国会議事堂前駅,*,*,国会議事堂前,,
        まで    助詞,副助詞,*,*,*,*,まで,マデ,マデ
        歩き    動詞,自立,*,*,五段・カ行イ音便,連用形,歩く,アルキ,アルキ
        まし    助動詞,*,*,*,特殊・マス,連用形,ます,マシ,マシ
        た      助動詞,*,*,*,特殊・タ,基本形,た,タ,タ
        。      記号,句点,*,*,*,*,。,。,。
        EOS
        <BLANKLINE>
        """
        self._check_initialized()
        return self.capi_ma.parse(sentence)

    def ma_parseNode(self, sentence):
        """
        センテンスを形態素解析した結果を MeCab 互換のノード配列として返します。

        Parameters
        ----------
        sentence : str
            解析する文字列。

        Returns
        -------
        list
            解析結果のリスト。

        Examples
        --------
        >>> from pygeonlp.api.service import Service
        >>> service = Service()
        >>> service.ma_parseNode('今日は国会議事堂前まで歩きました。')
        [{'conjugated_form': '*', 'conjugation_type': '*', 'original_form': '*', 'pos': 'BOS/EOS', 'prononciation': '*', 'subclass1': '*', 'subclass2': '*', 'subclass3': '*', 'surface': '', 'yomi': '*'}, {'conjugated_form': '*', 'conjugation_type': '*', 'original_form': '今日', 'pos': '名詞', 'prononciation': 'キョー', 'subclass1': '副詞可能', 'subclass2': '*', 'subclass3': '*', 'surface': '今日', 'yomi': 'キョウ'}, {'conjugated_form': '*', 'conjugation_type': '*', 'original_form': 'は', 'pos': '助詞', 'prononciation': 'ワ', 'subclass1': '係助詞', 'subclass2': '*', 'subclass3': '*', 'surface': 'は', 'yomi': 'ハ'}, {'conjugated_form': '*', 'conjugation_type': '*', 'original_form': '国会議事堂前', 'pos': '名詞', 'prononciation': '', 'subclass1': '固有名詞', 'subclass2': '地名語', 'subclass3': 'Bn4q6d:国会議事堂前駅/cE8W4w:国会議事堂前駅', 'surface': '国会議事堂前', 'yomi': ''}, {'conjugated_form': '*', 'conjugation_type': '*', 'original_form': 'まで', 'pos': '助詞', 'prononciation': 'マデ', 'subclass1': '副助詞', 'subclass2': '*', 'subclass3': '*', 'surface': 'まで', 'yomi': 'マデ'}, {'conjugated_form': '五段・カ行イ音便', 'conjugation_type': '連用形', 'original_form': '歩く', 'pos': '動詞', 'prononciation': 'アルキ', 'subclass1': '自立', 'subclass2': '*', 'subclass3': '*', 'surface': '歩き', 'yomi': 'アルキ'}, {'conjugated_form': '特殊・マス', 'conjugation_type': '連用形', 'original_form': 'ます', 'pos': '助動詞', 'prononciation': 'マシ', 'subclass1': '*', 'subclass2': '*', 'subclass3': '*', 'surface': 'まし', 'yomi': 'マシ'}, {'conjugated_form': '特殊・タ', 'conjugation_type': '基本形', 'original_form': 'た', 'pos': '助動詞', 'prononciation': 'タ', 'subclass1': '*', 'subclass2': '*', 'subclass3': '*', 'surface': 'た', 'yomi': 'タ'}, {'conjugated_form': '*', 'conjugation_type': '*', 'original_form': '。', 'pos': '記号', 'prononciation': '。', 'subclass1': '句点', 'subclass2': '*', 'subclass3': '*', 'surface': '。', 'yomi': '。'}, {'conjugated_form': '*', 'conjugation_type': '*', 'original_form': '*', 'pos': 'BOS/EOS', 'prononciation': '*', 'subclass1': '*', 'subclass2': '*', 'subclass3': '*', 'surface': '', 'yomi': '*'}]
        """
        self._check_initialized()
        return self.capi_ma.parseNode(sentence)

    def getWordInfo(self, geolod_id):
        """
        指定した geolod_id を持つ語の情報を返します。
        id が辞書に存在しない場合は None を返します。

        Parameters
        ----------
        geolod_id : str or Iterable
            語の ID または ID のリストを返すイテレータ。

        Returns
        -------
        dict
            geolod_id をキー、語の情報を値に持つ dict。

        Examples
        --------
        >>> from pygeonlp.api.service import Service
        >>> service = Service()
        >>> service.getWordInfo('Bn4q6d')
        {'body': '国会議事堂前', 'dictionary_id': 3, 'entry_id': 'LrGGxY', 'geolod_id': 'Bn4q6d', 'hypernym': ['東京地下鉄', '4号線丸ノ内線'], 'institution_type': '民営鉄道', 'latitude': '35.674845', 'longitude': '139.74534166666666', 'ne_class': '鉄道施設/鉄道駅', 'railway_class': '普通鉄道', 'suffix': ['駅', ''], 'dictionary_identifier': 'geonlp:ksj-station-N02'}
        """
        self._check_initialized()
        if isinstance(geolod_id, str):
            geoword = self.capi_ma.getWordInfo(geolod_id)
            return self._add_dict_identifier(geoword)

        results = {}
        for id in geolod_id:
            geoword = self.capi_ma.getWordInfo(id)
            if geoword is None:
                results[id] = None
            else:
                results[id] = self._add_dict_identifier(geoword)

        return results

    def searchWord(self, key):
        """
        指定した表記または読みを持つ語の情報を返します。
        一致する語が辞書に存在しない場合は None を返します。

        Parameters
        ----------
        key : str
            語の表記または読み。

        Returns
        -------
        dict
            geolod_id をキー、語の情報を値に持つ dict。

        Examples
        --------
        >>> from pygeonlp.api.service import Service
        >>> service = Service()
        >>> service.searchWord('国会議事堂前')
        {'Bn4q6d': {'body': '国会議事堂前', 'dictionary_id': 3, 'entry_id': 'LrGGxY', 'geolod_id': 'Bn4q6d', 'hypernym': ['東京地下鉄', '4号線丸ノ内線'], 'institution_type': '民営鉄道', 'latitude': '35.674845', 'longitude': '139.74534166666666', 'ne_class': '鉄道施設/鉄道駅', 'railway_class': '普通鉄道', 'suffix': ['駅', ''], 'dictionary_identifier': 'geonlp:ksj-station-N02'}, 'cE8W4w': {'body': '国会議事堂前', 'dictionary_id': 3, 'entry_id': '4NFELa', 'geolod_id': 'cE8W4w', 'hypernym': ['東京地下鉄', '9号線千代田線'], 'institution_type': '民営鉄道', 'latitude': '35.673543333333335', 'longitude': '139.74305333333334', 'ne_class': '鉄道施設/鉄道駅', 'railway_class': '普通鉄道', 'suffix': ['駅', ''], 'dictionary_identifier': 'geonlp:ksj-station-N02'}}
        """
        self._check_initialized()
        results = {}
        for k, w in self.capi_ma.searchWord(key).items():
            results[k] = self._add_dict_identifier(w)

        return results

    def getActiveDictionaries(self):
        """
        インストール済み辞書のうち、解析に利用する辞書のメタデータ一覧を返します。
        デフォルトでは全てのインストール済み辞書を利用します。

        一時的に辞書を利用したくない場合、 ``disactivateDictionaries()`` で除外できます。
        除外された辞書は ``activateDictionaries()`` で再び利用可能になります。

        Returns
        -------
        list
            Metadata インスタンスのリスト。

        Examples
        --------
        >>> from pygeonlp.api.service import Service
        >>> service = Service()
        >>> sorted([x.get_identifier() for x in service.getActiveDictionaries()])
        ['geonlp:geoshape-city', 'geonlp:geoshape-pref', 'geonlp:ksj-station-N02']
        """
        self._check_initialized()
        dictionaries = []
        for id, metadata in self.capi_ma.getActiveDictionaries().items():
            dictionaries.append(Metadata(metadata))

        return dictionaries

    def setActiveDictionaries(self, idlist=None, pattern=None):
        """
        インストール済み辞書のうち、解析に利用する辞書を指定します。

        Parameters
        ----------
        idlist : list, optional
            利用する辞書の id または identifier のリスト。
        pattern : str, optional
            利用する辞書の identifier の正規表現。

        Examples
        --------
        >>> from pygeonlp.api.service import Service
        >>> service = Service()
        >>> service.setActiveDictionaries(pattern=r'geonlp:geoshape')
        >>> sorted([x.get_identifier() for x in service.getActiveDictionaries()])
        ['geonlp:geoshape-city', 'geonlp:geoshape-pref']

        Note
        ----
        idlist と pattern のどちらかは指定する必要があります。
        """
        self._check_initialized()
        if idlist is not None and not isinstance(idlist, (list, set, tuple)):
            raise TypeError("idlist は None またはリストで指定してください。")

        if pattern is not None and not isinstance(pattern, str):
            raise TypeError("pattern は正規表現文字列で指定してください。")

        dictionaries = self.capi_ma.getDictionaryList()
        active_dictionaries = []
        for dic_id, dictionary in dictionaries.items():
            metadata = Metadata(dictionary)
            identifier = metadata.get_identifier()

            if idlist is not None and \
               (dic_id in idlist or identifier in idlist):
                logger.debug("id または identifier が idlist に含まれる")
                active_dictionaries.append(int(dic_id))
                continue

            if pattern is not None and re.search(pattern, identifier):
                logger.debug("identifier が pattern に一致")
                active_dictionaries.append(int(dic_id))
                continue

        if len(active_dictionaries) == 0:
            logger.debug("条件に一致する辞書がありません。")

        self.capi_ma.setActiveDictionaries(active_dictionaries)

    def disactivateDictionaries(self, idlist=None, pattern=None):
        """
        指定した辞書を一時的に除外し、利用しないようにします。
        既に除外されている辞書は除外されたままになります。
        新たに除外された辞書のリストを返します。

        Parameters
        ----------
        idlist : list
            除外する辞書の内部 id または identifier のリスト。
        pattern : str
            除外する辞書の identifier を指定する正規表現。

        Returns
        -------
        list
            新たに除外された辞書の identifier のリスト。

        Examples
        --------
        >>> from pygeonlp.api.service import Service
        >>> service = Service()
        >>> service.disactivateDictionaries(pattern=r'geonlp:geoshape')
        ['geonlp:geoshape-city', 'geonlp:geoshape-pref']
        >>> [x.get_identifier() for x in service.getActiveDictionaries()]
        ['geonlp:ksj-station-N02']
        >>> service.disactivateDictionaries(pattern=r'ksj-station')
        ['geonlp:ksj-station-N02']
        >>> [x.get_identifier() for x in service.getActiveDictionaries()]
        []

        Note
        ----
        idlist と pattern を同時に指定した場合、どちらか一方の条件に
        一致する辞書は除外されます。
        """
        self._check_initialized()
        if idlist is not None and not isinstance(idlist, (list, set, tuple)):
            raise TypeError("idlist は None またはリストで指定してください。")

        if pattern is not None and not isinstance(pattern, str):
            raise TypeError("pattern は正規表現文字列で指定してください。")

        current_active_dictionaries = self.capi_ma.getActiveDictionaries()
        new_active_dictionaries = []
        disactivated_dictionaries = []
        for dic_id, dictionary in current_active_dictionaries.items():
            metadata = Metadata(dictionary)
            identifier = metadata.get_identifier()

            if idlist is not None and \
               (dic_id in idlist or identifier in idlist):
                logger.debug("id または identifier が idlist に含まれる")
                disactivated_dictionaries.append(identifier)
                continue

            if pattern is not None and re.search(pattern, identifier):
                logger.debug("identifier が pattern に一致")
                disactivated_dictionaries.append(identifier)
                continue

            new_active_dictionaries.append(int(dic_id))

        if len(new_active_dictionaries) == 0:
            logger.debug("全ての辞書が除外されます。")

        self.capi_ma.setActiveDictionaries(new_active_dictionaries)

        return disactivated_dictionaries

    def activateDictionaries(self, idlist=None, pattern=None):
        """
        指定した辞書を再び利用するようにします。
        既に利用可能な辞書は指定しなくても利用可能なままになります。
        新たに利用可能になった辞書のリストを返します。

        Parameters
        ----------
        idlist : list
            利用する辞書の内部 id または identifier のリスト。
        pattern : str
            利用する辞書の identifier を指定する正規表現。

        Returns
        -------
        list
            新たに利用可能になった辞書の identifier のリスト。

        Examples
        --------
        >>> from pygeonlp.api.service import Service
        >>> service = Service()
        >>> sorted(service.disactivateDictionaries(pattern=r'.*'))
        ['geonlp:geoshape-city', 'geonlp:geoshape-pref', 'geonlp:ksj-station-N02']
        >>> [x.get_identifier() for x in service.getActiveDictionaries()]
        []
        >>> service.activateDictionaries(pattern=r'ksj-station')
        ['geonlp:ksj-station-N02']
        >>> [x.get_identifier() for x in service.getActiveDictionaries()]
        ['geonlp:ksj-station-N02']

        Note
        ----
        idlist と pattern を同時に指定した場合、どちらか一方の条件に
        一致する辞書は利用可能になります。
        """
        self._check_initialized()
        if idlist is not None and not isinstance(idlist, (list, set, tuple)):
            raise TypeError("idlist は None またはリストで指定してください。")

        if pattern is not None and not isinstance(pattern, str):
            raise TypeError("pattern は正規表現文字列で指定してください。")

        dictionaries = self.capi_ma.getDictionaryList()
        activated_dictionaries = []
        active_dictionaries = [
            int(x) for x in self.capi_ma.getActiveDictionaries().keys()
        ]
        for dic_id, dictionary in dictionaries.items():
            metadata = Metadata(dictionary)
            identifier = metadata.get_identifier()

            if int(dic_id) in active_dictionaries:
                # 既に有効
                continue

            if idlist is not None and \
               (dic_id in idlist or identifier in idlist):
                logger.debug("id または identifier が idlist に含まれる")
                active_dictionaries.append(int(dic_id))
                activated_dictionaries.append(identifier)
                continue

            if pattern is not None and re.search(pattern, identifier) and \
                    int(dic_id) not in active_dictionaries:
                logger.debug("identifier が pattern に一致")
                active_dictionaries.append(int(dic_id))
                activated_dictionaries.append(identifier)
                continue

        self.capi_ma.setActiveDictionaries(active_dictionaries)

        return activated_dictionaries

    def getActiveClasses(self):
        """
        解析に利用する固有名クラスの一覧を返します。
        デフォルトは '.*' で、全ての固有名クラスが利用されます。

        一時的に特定の固有名クラスだけを解析対象としたい場合、
        setActiveClasses() で対象クラスを指定できます。

        Returns
        -------
        list
            利用する固有名クラス（正規表現）のリスト。

        Examples
        --------
        >>> from pygeonlp.api.service import Service
        >>> service = Service()
        >>> service.getActiveClasses()
        ['.*']
        """
        self._check_initialized()
        return self.capi_ma.getActiveClasses()

    def setActiveClasses(self, patterns=None):
        """
        解析対象とする固有名クラスの正規表現リストを指定します。
        いずれかの正規表現に一致する固有名クラスは解析対象となります。

        '-' から始まる場合、その正規表現に一致する固有名クラスは対象外となります。

        Parameters
        ----------
        patterns : list, optional
            解析対象とする固有名クラス（str）のリスト。
            省略した場合 ['.*'] （全固有名クラス）を対象とします。

        Examples
        --------
        >>> from pygeonlp.api.service import Service
        >>> service = Service()
        >>> service.getActiveClasses()
        ['.*']
        >>> service.searchWord('東京都')
        {'QknGsa': {...'dictionary_identifier': 'geonlp:geoshape-pref'}}
        >>> service.setActiveClasses(['.*', '-都道府県'])
        >>> service.searchWord('東京都')
        {}
        >>> service.setActiveClasses()
        >>> service.getActiveClasses()
        ['.*']
        """
        self._check_initialized()
        if patterns is None:
            patterns = ['.*']
        elif isinstance(patterns, str):
            patterns = [patterns]

        self.capi_ma.setActiveClasses(patterns)

    def _check_initialized(self):
        """
        capi オブジェクトが初期化されていることを確認します。
        """
        if self.capi_ma is None:
            raise ServiceError("Service が初期化されていません。")

    def _add_dict_identifier(self, word):
        """
        語に含まれる辞書の内部 ID "dictionary_id" を見て
        語に辞書の識別子 "dictionary_identifier" を追加します。
        """
        if word is None or 'dictionary_id' not in word or \
                'dictionary_identifier' in word:
            return word

        dictionary_id = word['dictionary_id']
        if dictionary_id in self._dict_cache:
            identifier = self._dict_cache[dictionary_id]
        else:
            identifier = self.capi_ma.getDictionaryIdentifierById(
                dictionary_id)

            self._dict_cache[dictionary_id] = identifier

        word['dictionary_identifier'] = identifier
        return word



"""
コンテキスト情報を利用した geoparsing のサンプル

地理的範囲と固有名クラスごとの優先度をもつ「コンテキスト情報」を
利用するカスタムクラスの実装例です。

Usage: python context.py

以下のクラスを実装し、サンプルテキストを解析した結果を表示します。

- コンテキスト情報の範囲内かつ優先度０以上の固有名クラスを持つ
  地名語だけを残す ContextFilter
- 固有名クラスに応じてスコアを乗じる ContextScoringClass
- ContextScoringClass を利用する ContextEvaluator
- ContextFilter, ContextEvaluator を利用する ContextWorkflow
"""

from logging import getLogger
import os
import re
from typing import Optional, Union

from pygeonlp.api.linker import Evaluator
from pygeonlp.api.filter import Filter
from pygeonlp.api.node import Node
from pygeonlp.api.scoring import ScoringClass
from pygeonlp.api.spatial_filter import GeoContainsFilter
from pygeonlp.api.workflow import Workflow

logger = getLogger(__name__)


class Context(object):
    """
    コンテキスト情報を保持するクラス

    Attributes
    ----------
    target_area: str, bytes, PathLike, None
        このコンテキストが対象とする空間的範囲を表す GeoJSON を
        含むファイルパスまたは URL
    priorities: dict
        固有名クラスごとの優先度
    priorities_with_prefix: dict
        固有名クラスごとの優先度（prefix を含む場合）
    priorities_with_suffix: dict
        固有名クラスごとの優先度（suffix を含む場合）
    """

    def __init__(self):
        """
        コンテキスト情報を初期化
        """
        self.target_area = None
        self.priorities = {}
        self.priorities_with_prefix = {}
        self.priorities_with_suffix = {}

    def setArea(self, geojson_or_url: Union[str, bytes, os.PathLike]):
        """
        空間的範囲をセットする

        Parameters
        ----------
        geojson_or_url: str, bytes, PathLike
            空間的範囲を表す GeoJSON または
            GeoJSON を含む URL
        """
        self.target_area = geojson_or_url

    def getPriority(self, node: Node) -> float:
        """
        指定した形態素ノードに対応する
        このコンテキストでの優先度を返す

        Parameters
        ----------
        node: pygeonlp.api.node.Node
            形態素ノード

        Return
        ------
        float
            優先度
        """
        if node.node_type == Node.ADDRESS:
            # 住所は常に優先度最高 = 1
            return 1.0

        if node.node_type == Node.NORMAL:
            # 非地名語は常に優先度最低 = 0
            return 0.0

        re_ne_class = re.compile(node.prop.get('ne_class'))

        # ノードが prefix, suffix を含むかどうかを判定
        body_pos = node.surface.index(node.prop.get('body'))
        has_prefix = body_pos > 0
        has_suffix = (body_pos +
                      len(node.prop.get('body'))) < len(node.surface)

        if has_suffix:
            # suffix を含む場合の ne_class の優先度を取得
            for ne_class, priority in self.priorities_with_suffix.items():
                if re_ne_class.match(ne_class):
                    return priority

        if has_prefix:
            # prefix を含む場合の ne_class の優先度を取得
            for ne_class, priority in self.priorities_with_prefix.items():
                if re_ne_class.match(ne_class):
                    return priority

        # prefix, suffix を含まない場合の ne_class の優先度を取得
        for ne_class, priority in self.priorities.items():
            if re_ne_class.match(ne_class):
                return priority

        # 設定されていない場合は 0.5
        return 0.5


class ContextFilter(Filter):
    """
    コンテキスト情報を利用して地名語候補を選択するフィルタクラス

    Attributes
    ----------
    context: Context
        利用するコンテキスト情報
    """

    def __init__(self, context: Optional[Context] = None):
        """
        Note
        ----
        このフィルタは、一致する候補が見つからない場合は
        非地名語の固有名詞として処理します。
        """
        self.when_all_failed = 'convert_to_normal'
        self.set_context(context)

    def set_context(self, context: Optional[Context] = None):
        """
        コンテキスト情報をセットする

        Parameters
        ----------
        context: Context, optional
            利用するコンテキスト情報。

        Note
        ----
        コンテキスト情報が空間範囲を持っている場合、
        範囲内の候補を選択するための GeoContainsFilter を作成し、
        self._filter に保持します。
        """
        self.context = context

        if self.context and self.context.target_area:
            self._filter = GeoContainsFilter(
                self.context.target_area)
        else:
            self._filter = None

    def apply_filter(self, candidates, **kwargs):
        """
        Filter.apply_filter() をオーバーライド

        ラティス表現の1つの形態素に対する候補ノード集合
        candidates のうち、コンテキストの空間範囲に含まれ、
        かつ優先度が定義されている地名語クラスを持つノードを
        残します。それ以外は除去します。

        Parameters
        ----------
        candidates : list[Node]
            候補ノードのリスト。
        **kwargs
            キーワードパラメータ。このクラスでは利用しません。

        Returns
        -------
        list[Node]
            条件に一致した候補ノードのリスト。
        """
        if self._filter:
            # 空間範囲が指定されていれば絞り込む
            filtered_candidates = []
            for candidate in candidates:
                if self._filter.filter_func(candidate):
                    filtered_candidates.append(candidate)
        else:
            filtered_candidates = candidates

        # 優先度がセットされている候補だけを含むリストを作成する
        results = []
        for candidate in filtered_candidates:
            priority = self.context.getPriority(node=candidate)
            if priority > 0.0:
                results.append(candidate)

        return results


class ContextScoringClass(ScoringClass):
    """
    コンテキスト情報を利用してパスのスコアを計算する
    スコアリングクラス

    パスに地名語が含まれている場合、コンテキスト情報で
    定義されている地名語クラスの優先度を乗算します。
    """

    def __init__(self, context: Context):
        super().__init__()
        self.context = context

    def path_score(self, path):
        """
        デフォルト実装のスコアにコンテキスト情報の
        優先度を乗算します。
        """
        score = super().path_score(path)

        geowords = Evaluator.collect_geowords(path)
        for geoword in geowords:
            score *= self.context.getPriority(geoword)
            if score == 0.0:
                break

        return score


class ContextEvaluator(Evaluator):
    """
    コンテキスト情報を利用して最適な候補を返す Evaluator
    """

    def __init__(self, context: Context, max_results: int = 5):
        super().__init__(
            scoring_class=ContextScoringClass(context),
            max_results=max_results)


class ContextWorkflow(Workflow):
    """
    コンテキスト情報を利用してテキスト解析を行なう Workflow
    """

    def __init__(self, context,
                 db_dir=None, address_regex=None, jageocoder=None,
                 max_results=5):
        self.max_results = max_results
        super().__init__(
            db_dir=db_dir,
            address_regex=address_regex,
            jageocoder=jageocoder)

        self.set_context(context)

    def set_context(self, context):
        """
        利用するコンテキスト情報をセットします

        Parameters
        ----------
        context: Context
            利用するコンテキスト情報

        Note
        ----
        セットされたコンテキスト情報を利用して、
        filters と evaluator を作り直します。
        """
        self.context = context
        self.filters = [ContextFilter(context)]
        self.evaluator = ContextEvaluator(context, self.max_results)


if __name__ == '__main__':
    import logging
    from pygeonlp.api.workflow import Workflow
    from pygeonlp.api.devtool import pp_geojson

    logging.basicConfig(level=logging.WARNING)
    with open(os.path.join(
            os.path.dirname(__file__),
            'nihonkai_engan.geojson'), 'r') as f:
        geojson = f.read()

    context_heavy_snowfall_20220113 = Context()
    context_heavy_snowfall_20220113.setArea(geojson)
    context_heavy_snowfall_20220113.priorities = {
        r'道路施設/バス停': 0.0,
        r'道路施設/.*': 0.7,
        r'鉄道施設/.*': 0.6,
        r'市区町村': 0.8,
    }
    context_heavy_snowfall_20220113.priorities_with_suffix = {
        r'道路施設/.*': 1.0,
        r'鉄道施設/.*': 1.0,
        r'市区町村': 1.0,
    }

    workflow = ContextWorkflow(
        context_heavy_snowfall_20220113,
        max_results=1)

    article = """北海道 NEWS WEB
https://www3.nhk.or.jp/sapporo-news/20220113/7000042131.html

大雪の影響続く ＪＲ各線で計１９１本が運休

01月13日　11時29分

ＪＲ北海道は、１３日も大雪の影響で一部の特急や普通列車などあわせて１９１本の運休を決めています。

ＪＲ北海道によりますと、１３日も大雪の影響で４つの在来線で運転の見合わせが出ていて、午前１１時現在であわせて１９１本が運休、または運休を決めています。
▼室蘭線は、岩見沢と苫小牧の間で始発から午後３時ごろまで、▼宗谷線は、名寄と南稚内の間で始発から午後２時半ごろまで、南稚内と稚内の間で始発から終日、運転を見合わせます。
また▼石勝線は、南千歳と新夕張の間で午後５時ごろまで、▼釧網線も、釧路と網走の間で午後２時ごろまでそれぞれ始発から運転を見合わせます。
１３日、これまでに運休が決まっている特急は、札幌と釧路を結ぶ「おおぞら」や稚内と旭川を結ぶ「サロベツ」など合わせて１７本に上っています。
このほか在来線では▼函館・千歳線が札幌と新千歳空港を結ぶ快速エアポート３５本などあわせて６３本、▼函館線で３８本、▼根室線で２０本、▼室蘭線で１２本、▼釧網線で９本、▼宗谷線で９本、▼石北線で８本、▼石勝線で８本、▼学園都市線で７本が運休となっています。
ＪＲ北海道は、今後の天候によってはさらに運休が増えるおそれもあるとしてホームページで最新の情報を確認するよう呼びかけています。
"""
    # 結果の GeoJSON を pp_geojson() で簡易表示します
    pp_geojson(workflow.geoparse(article))
    # GeoJSON で出力するには json.dumps() を利用します
    # import json
    # print(json.dumps(
    #     workflow.geoparse(article),
    #     indent=2, ensure_ascii=False))



from logging import getLogger
import os

from pygeonlp.api.node import Node

logger = getLogger(__name__)


MAX_COMBINATIONS = int(os.environ.get("GEONLP_MAX_COMBINATIONS", "256"))


class LinkerError(RuntimeError):
    pass


class LinkedResults(object):
    """
    ラティス表現からパス表現の候補を
    次々に出力するイテレータを作成します。

    Examples
    --------
    >>> import pygeonlp.api as api
    >>> from pygeonlp.api.linker import LinkedResults
    >>> api.init()
    >>> for lr in LinkedResults(api.analyze('国会議事堂前まで歩きました。')):
    ...     [x.simple() for x in lr]
    ["国会議事堂前(GEOWORD:['東京地下鉄', '4号線丸ノ内線'])", 'まで(NORMAL)', '歩き(NORMAL)', 'まし(NORMAL)', 'た(NORMAL)', '。(NORMAL)']
    ["国会議事堂前(GEOWORD:['東京地下鉄', '9号線千代田線'])", 'まで(NORMAL)', '歩き(NORMAL)', 'まし(NORMAL)', 'た(NORMAL)', '。(NORMAL)']
    """

    def __init__(self, lattice):
        """
        Parameters
        ----------
        lattice : list
            入力となるラティス表現

        Returns
        -------
        int
            常に 0
        """
        self.lattice = lattice
        self.reset_counter()

    def __iter__(self):
        return self

    def __next__(self):
        """
        カウンターが示している候補リストを返し、カウンターを次に進めます。
        """
        output = self.get_result()
        if output is None:
            raise StopIteration()

        self.increment_counter()
        return output

    def counter(self):
        """
        カウンターを返します。
        """
        return self.counters

    def get_result(self):
        """
        カウンターが示している候補リストを返します。
        カウンターは次に進めません。
        """
        if self.counters[0] == -1:
            # 既に最後まで取得済みの場合は None を返す
            return None

        output = []
        n = 0
        while n < len(self.lattice):
            node = self.lattice[n][self.counters[n]]
            output.append(node)

            # 住所ノード以外が保存されているかチェック
            has_non_address_node = False
            for _node in self.lattice[n]:
                if _node.node_type != Node.ADDRESS:
                    has_non_address_node = True
                    break

            if isinstance(node.morphemes, (list, tuple,)) and \
                    has_non_address_node:
                n += len(node.morphemes)
            else:
                n += 1

        return output

    def reset_counter(self):
        """
        カウンターを初期化します。
        """
        self.counters = [0] * len(self.lattice)
        # 形態素ノードの種別を事前に計算する
        # 0: 非地名語のみ
        # 1: 地名語を含む（住所は含まない）
        # 2: 住所の先頭ノード
        self.node_types = [Node.NORMAL] * len(self.lattice)
        n = 0
        while n < len(self.lattice):
            nodes = self.lattice[n]

            for node in nodes:
                if node.node_type == Node.ADDRESS:
                    self.node_types[n] = Node.ADDRESS
                    break
                elif node.node_type == Node.GEOWORD:
                    self.node_types[n] = Node.GEOWORD

            n += 1

        logger.debug("ノード種別リスト: {}".format(self.node_types))

    def increment_counter(self):
        """
        カウンターを次に進めます。
        """
        positions = []
        pos = 0

        # ラティス上で地名語・住所ノード候補を含む位置のリストを作成
        # 住所ノードが選択されている場合は住所に含まれる要素はスキップ
        while pos < len(self.lattice):
            i = self.counters[pos]
            node = self.lattice[pos][i]
            next_pos = pos + 1

            if node.node_type == Node.ADDRESS:
                positions.append(pos)
                has_non_address_node = False
                for _node in self.lattice[pos]:
                    if _node.node_type != Node.ADDRESS:
                        has_non_address_node = True
                        break

                if has_non_address_node:
                    # この形態素には住所ノード以外も存在するので
                    # 住所が終わる位置までスキップ
                    next_pos = pos + len(node.morphemes)

            elif node.node_type == Node.GEOWORD:
                positions.append(pos)

            pos = next_pos

        # 位置のリストの最後尾から、カウンターを 1 増やす
        # 候補の数を超えた場合は 0 に戻し、
        # 一つ手前の位置でインクリメントを継続する
        while len(positions) > 0:
            pos = positions.pop()
            i = self.counters[pos] + 1
            if i < len(self.lattice[pos]):
                self.counters[pos] = i
                return
            else:
                self.counters[pos] = 0

        # インクリメントできなかった = 最後の候補に到達した場合
        # カウンターの先頭を -1 にセット
        self.counters[0] = -1


class Evaluator(object):
    """
    ラティス表現から、指定したメソッドで計算したスコアの高いものから
    順に並べたパス表現を作成します。

    スコアリングに独自のメソッドを利用する場合は ``scoring_method`` で指定してください。

    Examples
    --------
    >>> import pygeonlp.api as api
    >>> from pygeonlp.api.linker import Evaluator
    >>> api.init()
    >>> rr = Evaluator(max_results=5)
    >>> for x in rr.get(api.analyze('福島は大阪から2分です。')):
    ...   (x['score'], [n.simple() for n in x['result']])
    ...
    (46, ["福島(GEOWORD:['西日本旅客鉄道', '大阪環状線'])", 'は(NORMAL)', "大阪(GEOWORD:['西日本旅客鉄道', '大阪環状線'])", 'から(NORMAL)', '2(NORMAL)', '分(NORMAL)', 'です(NORMAL)', '。(NORMAL)'])
    (46, ["福島(GEOWORD:['西日本旅客鉄道', '大阪環状線'])", 'は(NORMAL)', "大阪(GEOWORD:['西日本旅客鉄道', '東海道線'])", 'から(NORMAL)', '2(NORMAL)', '分(NORMAL)', 'です(NORMAL)', '。(NORMAL)'])
    (41, ["福島(GEOWORD:['阪神電気鉄道', '本線'])", 'は(NORMAL)', "大阪(GEOWORD:['西日本旅客鉄道', '大阪環状線'])", 'から(NORMAL)', '2(NORMAL)', '分(NORMAL)', 'です(NORMAL)', '。(NORMAL)'])
    (41, ["福島(GEOWORD:['阪神電気鉄道', '本線'])", 'は(NORMAL)', "大阪(GEOWORD:['西日本旅客鉄道', '東海道線'])", 'から(NORMAL)', '2(NORMAL)', '分(NORMAL)', 'です(NORMAL)', '。(NORMAL)'])
    (36, ['福島(GEOWORD,福島県)', 'は(NORMAL)', '大阪(GEOWORD,大阪府)', 'から(NORMAL)', '2(NORMAL)', '分(NORMAL)', 'です(NORMAL)', '。(NORMAL)'])

    Attributes
    ----------
    scoring_class : class instance
        スコアリングを行なうクラス。
    scoring_options : any
        スコアリングクラスの初期化に渡すオプションパラメータ。
    scorer : service.ScoringClass instance
        スコアリングを行なうクラスインスタンス。
    max_results : int
        保持する結果の最大数。
    """

    def __init__(self, scoring_class=None, scoring_options=None,
                 max_results=5, max_combinations=None):
        """
        Parameters
        ----------
        scoring_class : class, optional
            パスのスコアとノード間のスコアを計算する関数を持つ
            スコアリングクラス。
            指定しない場合、``pygeonlp.api.scoring`` モジュール内の
            ``ScoringClass`` が利用されます。
        scoring_options : any, optional
            スコアリングクラスの初期化に渡すオプションパラメータ。
        max_results : int, optional
            保持する結果の最大数を指定します（デフォルト = 5）。
        max_combinations : int, optional
            ノード候補の組み合わせ数の上限値。これを超える組み合わせが
            可能な入力が与えられた場合は例外 LinkerError を発生しますので
            必要に応じて大きくしてください。
            デフォルト値は linker.MAX_COMBINATIONS です。
        """
        from .scoring import ScoringClass
        self.scoring_class = scoring_class
        self.scoring_options = scoring_options
        self.max_results = max_results
        self.max_combinations = max_combinations 

        if self.scoring_class is None:
            self.scorer = ScoringClass(scoring_options)
        elif isinstance(self.scoring_class, ScoringClass):
            self.scorer = self.scoring_class
        else:
            self.scorer = scoring_class(scoring_options)

        if self.max_combinations is None:
            self.max_combinations = MAX_COMBINATIONS

    def count_combinations(self, lattice):
        """
        ラティス形式の入力に対し、組み合わせたパス表現の個数を計算します。

        Parameters
        ----------
        lattice : list
            入力となるラティス表現。

        Return
        ------
        int
            組み合わせの数。
        """
        n = 1
        for i in lattice:
            n *= len(i)
            if n > 2147483647:
                break

        return n

    def get(self, lattice):
        """
        ラティス表現を入力として、スコアリングと並べ替えを行ないます。

        Parameters
        ----------
        lattice : list
            入力となるラティス表現。

        Return
        ------
        list
            スコアを 'score', パス表現の解析結果を 'result' に持つ
            dict のリスト。
            スコア降順にソートされ、最大 max_results 個の要素を含みます。
        """
        results = []
        combination = self.count_combinations(lattice)
        if combination > self.max_combinations:
            raise LinkerError(
                "組み合わせ数 {} がしきい値 {} を超えています。".format(
                    combination, self.max_combinations))

        lr = LinkedResults(lattice)
        for path in lr:
            score = self.scorer.path_score(path)
            new_record = {
                "score": score,
                "result": path,
            }
            simple_repr = ''.join([x.simple() for x in path])
            logger.debug("{} => {}".format(simple_repr, score))
            logger.debug("next:{}".format(lr.counter()))

            # スコアが高いものから順に最大 max_results 個の結果を保持する
            n = len(results)
            if n == 0:
                results.append(new_record)
                continue

            while n > 0:
                if results[n - 1]['score'] >= score:
                    results.insert(n, new_record)
                    break

                n -= 1
                if n == 0:
                    results.insert(n, new_record)

            results = results[0:self.max_results]

        return results

    def as_dict(self, lattice):
        """
        ``get()`` と同じ処理を行ないますが、結果に含まれるノードの情報を
        JSON に変換可能な dict に変換してから返します。

        Parameters
        ----------
        lattice : list
            入力となるラティス表現。

        Return
        ------
        list
            ``get()`` の出力結果を JSON 変換可能な形式に変換したリスト。
        """
        results = []
        for r in self.get(lattice):
            score = r['score']
            node_list = r['result']
            dict_list = [x.as_dict() for x in node_list]
            results.append({
                "score": score,
                "result": dict_list,
            })

        return results

    def as_geojson(self, lattice):
        """
        ``get()`` と同じ処理を行ないますが、結果に含まれるノードの情報を
        GeoJSON FeatureCollection に変換可能な dict に変換してから返します。

        Parameters
        ----------
        lattice : list
            入力となるラティス表現。

        Return
        ------
        list
            ``get()`` の出力結果を GeoJSON 変換可能な形式に変換したリスト。
        """
        results = []
        for r in self.get(lattice):
            score = r['score']
            node_list = r['result']
            features = [x.as_geojson() for x in node_list]
            results.append({
                "score": score,
                "geojson": {
                    "type": "FeatureCollection",
                    "features": features,
                },
            })

        return results

    @staticmethod
    def collect_geowords(result):
        """
        パス表現の結果に含まれる地名語のセットを返します。

        処理結果に含まれる地名語だけを列挙する場合に利用する簡易メソッドです。
        """
        geowords = []
        for node in result:
            if node.node_type == Node.GEOWORD:
                geowords.append(node)

        return geowords

    @staticmethod
    def collect_addresses(result):
        """
        パス表現の結果に含まれる住所セットを返します。

        処理結果に含まれる住所だけを列挙する場合に利用する簡易メソッドです。
        """
        addresses = []
        for node in result:
            if node.node_type == Node.ADDRESS:
                addresses.append(node)

        return addresses



import logging
from multiprocessing import Process
import time
import unittest

from pygeonlp.api.node import Node
import pygeonlp.api as api

logger = logging.getLogger(__name__)
# logging.basicConfig(level=logging.INFO)

"""
複数プロセスで同時アクセスを行なうテスト。

このモジュールのテストを行なうには、次のコマンドを実行してください。

python -m unittest -v pygeonlp.tests.test_multiprocess
"""


def analyze(query):
    service = api.service.Service()
    parser = api.parser.Parser(service=service)
    while True:
        parser.analyze_sentence(query)


class TestMultiProcessing(unittest.TestCase):

    queries = [
        'NIIは千代田区一ツ橋2-1-2にあります。',
        ('東京都恩賜上野動物公園は東京都台東区上野公園の'
            '上野恩賜公園内に在る東京都立動物園。'),
        ('北海道大学（ほっかいどうだいがく、英語: Hokkaido University）は、'
            '札幌市北区北8西5に本部を置く日本の国立大学である。'),
        '2013年9月の第125次IOC総会で開催都市が東京に決定した。',
    ]

    def setUp(self):
        api.init()
        self.parser = api.parser.Parser()

    def test_single_process(self):
        """
        Test analyze in single process.

        "analyze_sentence" should return the result as follows:
        ['NII(NORMAL)']
        ['は(NORMAL)']
        ["千代田区(GEOWORD:['東京都'])"]
        ['一ツ橋(NORMAL)']
        ['2(NORMAL)']
        ['-(NORMAL)']
        ['1(NORMAL)']
        ['-(NORMAL)']
        ['2(NORMAL)']
        ['に(NORMAL)']
        ['あり(NORMAL)']
        ['ます(NORMAL)']
        ['。(NORMAL)']
        """
        query = self.__class__.queries[0]
        lattice = self.parser.analyze_sentence(query)

        self.assertEqual(len(lattice), 13)

        nodes = lattice[2]
        self.assertEqual(len(nodes), 1)
        self.assertEqual(nodes[0].node_type, Node.GEOWORD)

    def test_five_processes(self):
        """
        Test simultaneous analyzes in multi-processes.
        """

        processes = []
        for q in self.__class__.queries:
            p = Process(target=analyze, args=(q, ))
            p.start()
            processes.append(p)

        time.sleep(10)

        for p in processes:
            p.terminate()


if __name__ == '__main__':
    unittest.main()



import pytest

import pygeonlp.api as geonlp_api
from pygeonlp.api.dict_manager import DictManager
from pygeonlp.webapi.app import app
from pygeonlp.webapi import config


@pytest.fixture(scope="session", autouse=True)
def createTempDatabase(tmp_path_factory):
    """
    テスト用に一時的な GeoNLP データベースを作成します。
    テストを実行する際に一度だけ実行されます。
    """
    if True:
        # Setup test database in a temporary directory
        db_dir = str(tmp_path_factory.mktemp('db'))
        manager = DictManager(db_dir=db_dir)
        manager.setupBasicDatabase()
        geonlp_api.init(db_dir=db_dir,
                        **(config.GEONLP_API_OPTIONS))
    else:
        # or, use default database under GEONLP_DB_DIR
        pass


@pytest.fixture(scope="function", autouse=True)
def setup():
    """
    各テスト実行前に自動的に行なう処理を呼びだします。
    """
    app.config['TESTING'] = True


@pytest.fixture
def client():
    """
    Flask アプリケーションのテストクライアントを返します。
    """
    with app.test_client() as client:
        with app.app_context():
            yield client


@pytest.fixture(scope="session")
def require_gdal():
    try:
        from osgeo import gdal
        return gdal.VersionInfo()
    except ModuleNotFoundError:
        pytest.skip("Skip (GDAL not installed)")



import json
import logging
import os
import site
import sys

from docopt import docopt
import pygeonlp.api
from pygeonlp.api.dict_manager import DictManager
from pygeonlp.api.devtool import pp_mecab

logger = logging.getLogger(__name__)

HELP = """
'pygeonlp' は、自然言語テキストを解析して地名を抽出する
ジオタギング (geo-tagging) またはジオパージング (geo-parsing) と呼ばれる
処理を行なうオープンソースソフトウェアです。

Usage:
  {p} -h
  {p} --version
  {p} geoparse [--db-dir=<dir>] [--json] [<file>]
  {p} search [--db-dir=<dir>] <word>
  {p} list-dictionaries [--db-dir=<dir>]
  {p} show-dictionary [--db-dir=<dir>] <id>
  {p} clear-dictionaries [--db-dir=<dir>]
  {p} add-dictionary [--db-dir=<dir>] (<json-path> <csv-path> | <url>)
  {p} remove-dictionary [--db-dir=<dir>] <id>
  {p} download-dictionary [--db-dir=<dir>] <url> <json-path> <csv-path>
  {p} setup [--db-dir=<dir>] [<dict-src-dir>]

Options:
  -h --help           このヘルプを表示します。
  --json              JSON で出力します。
  --db-dir=<dir>      データベースディレクトリを指定します。

Examples:
- "test.txt" のテキストをジオパーズした結果を出力します
  {p} geoparse test.txt

- "神保町" をデータベースから検索します
  {p} search 神保町

- パッケージに同梱されている辞書ファイルから初期データベースを設定します
  {p} setup

- インストールされている辞書情報一覧を表示します
  {p} list-dictionaries

- 指定した辞書の情報を表示します（辞書の識別子で指定）
  {p} show-dictionary geonlp:geoshape-city

- インストールされている辞書をクリア（全て削除）します
  {p} clear-dictionaries

- ウェブから辞書をインストールします
  {p} add-dictionary https://geonlp.ex.nii.ac.jp/dictionary/ksj-post-office/

- 指定した辞書をアンインストールします

  {p} remove-dictionary geonlp:post-office

- ウェブから辞書をダウンロードし、JSONおよびCSVファイルに保存します
  {p} download-dictionary https://geonlp.ex.nii.ac.jp/dictionary/geoshape-city/ geoshape.json geoshape.csv

- JSONおよびCSVファイルから辞書をインストールします
  {p} add-dictionary base_data/geoshape-city.json base_data/geoshape-city.csv

""".format(p='pygeonlp')


def setup_basic_database(db_dir=None, src_dir=None):
    """
    基本的な地名解析辞書を登録したデータベースを作成します。

    Parameters
    ----------
    db_dir : str, optional
        データベースディレクトリを指定します。
        ここにデータベースファイルを作成します。
        既にデータベースが存在する場合は追記します（つまり、
        既に登録済みの地名解析辞書のデータは失われません）。
        省略された場合には ``get_db_dir()`` で決定します。

    src_dir : str, optional
        地名解析辞書ファイルが配置されているディレクトリ。
        省略された場合には、 ``sys.prefix`` または ``site.USER_BASE`` の
        下に ``pygeonlp_basedata`` がないか探します。
        見つからない場合は ``RuntimeError`` を送出しますので、
        ディレクトリを指定してください。
    """
    data_dir = None
    base_dir = os.getcwd()
    candidates = []
    if src_dir:
        candidates.append(src_dir)

    candidates += [
        os.path.join(sys.prefix, 'pygeonlp_basedata'),
        os.path.join(site.USER_BASE, 'pygeonlp_basedata'),
        os.path.join(base_dir, 'base_data'),
        os.path.join(base_dir, 'pygeonlp_basedata'),
    ]

    for cand in candidates:
        if os.path.exists(cand):
            data_dir = cand
            break

    if not data_dir:
        files = DictManager.get_package_files()
        for path in files:
            pos = path.find('/pygeonlp_basedata')
            if pos < 0:
                continue

            data_dir = path[0:pos + len('/pygeonlp_basedata')]

    if not data_dir:
        raise RuntimeError("地名解析辞書がインストールされたディレクトリが見つかりません。")

    if db_dir is None:
        db_dir = pygeonlp.api.get_db_dir()

    os.makedirs(db_dir, 0o755, exist_ok=True)
    manager = DictManager(db_dir=db_dir)

    updated = False
    if manager.getDictionary('geonlp:geoshape-city') is None:
        manager.addDictionaryFromFile(
            jsonfile=os.path.join(data_dir, 'geoshape-city.json'),
            csvfile=os.path.join(data_dir, 'geoshape-city.csv'))
        updated = True

    if manager.getDictionary('geonlp:geoshape-pref') is None:
        manager.addDictionaryFromFile(
            jsonfile=os.path.join(data_dir, 'geoshape-pref.json'),
            csvfile=os.path.join(data_dir, 'geoshape-pref.csv'))
        updated = True

    if manager.getDictionary('geonlp:ksj-station-N02') is None:
        manager.addDictionaryFromFile(
            jsonfile=os.path.join(data_dir, 'ksj-station-N02.json'),
            csvfile=os.path.join(data_dir, 'ksj-station-N02.csv'))
        updated = True

    if updated:
        manager.updateIndex()


def main():
    args = docopt(HELP)

    if args['--version']:
        print(pygeonlp.api.__version__)
        exit(0)

    if args['--db-dir']:
        db_dir = args['--db-dir']
    else:
        db_dir = pygeonlp.api.get_db_dir()

    manager = pygeonlp.api.dict_manager.DictManager(db_dir)
    if manager.capi_ma is None and args['setup'] is False:
        while True:
            r = input(
                f"{db_dir} に基本辞書セットをインストールしますか? (y/n):"
            )
            if r[0].lower() == 'y':
                manager.setupBasicDatabase()
                print("完了しました。")
                break
            else:
                print((
                    "pygeonlp setup コマンドを実行して"
                    "辞書をインストールしてください。"
                ))
                exit(1)

    if args['geoparse']:
        pygeonlp.api.init(db_dir=db_dir)
        if args['<file>']:
            with open(args['<file>'], "r") as f:
                for line in f:
                    line = line.rstrip()
                    if line == "":
                        continue

                    geojson = pygeonlp.api.geoparse(line)
                    if args['--json']:
                        print(json.dumps(geojson, ensure_ascii=False))
                    else:
                        print(pp_mecab(geojson))

        else:
            for line in sys.stdin:
                line = line.rstrip()
                if line != "":
                    geojson = pygeonlp.api.geoparse(line)
                    if args['--json']:
                        print(json.dumps(geojson, ensure_ascii=False))
                    else:
                        pp_mecab(geojson)

        exit(0)

    if args['search']:
        pygeonlp.api.init(db_dir=db_dir)
        word = pygeonlp.api.searchWord(args["<word>"])
        print(json.dumps(word, ensure_ascii=False))
        exit(0)

    if args['list-dictionaries']:
        for metadata in manager.getDictionaries():
            dic = json.loads(metadata.jsonld)
            print("{} : {}\n- {}\n{}\n".format(
                metadata.get_identifier(),
                metadata.get_name(),
                dic["url"],
                dic["description"]))

        exit(0)

    if args['show-dictionary']:
        print(
            json.dumps(
                json.loads(manager.getDictionary(args['<id>']).jsonld),
                indent=2,
                ensure_ascii=False
            )
        )
        exit(0)

    if args['clear-dictionaries']:
        manager.clearDatabase()
        manager.updateIndex()
        exit(0)

    if args['add-dictionary']:
        if args['<json-path>']:
            manager.addDictionaryFromFile(
                args['<json-path>'], args['<csv-path>'])
        else:
            manager.addDictionaryFromWeb(args['<url>'])

        manager.updateIndex()
        exit(0)

    if args['remove-dictionary']:
        metadata = manager.getDictionary(args['<id>'])
        if metadata is not None:
            manager.removeDictionary(args['<id>'])
            manager.updateIndex()
        else:  # Supports cases specified by URLs.
            for metadata in manager.getDictionaries():
                dic = json.loads(metadata.jsonld)
                if dic["url"] == args["<id>"]:
                    identifier = metadata.get_identifier()
                    manager.removeDictionary(identifier)
                    manager.updateIndex()
                    break

            else:  # Call the method to raise error.
                manager.removeDictionary(args['<id>'])

        exit(0)

    if args['download-dictionary']:
        manager.saveDictionaryFromWeb(
            jsonfile=args['<json-path>'],
            csvfile=args['<csv-path>'],
            url=args['<url>'])

        exit(0)

    if args['setup']:
        src_dir = args['<dict-src-dir>']
        manager.setupBasicDatabase(src_dir=src_dir)
        exit(0)

    raise RuntimeError('Unexpected args: {}'.format(args))


if __name__ == '__main__':
    main()



import logging
import os
import unittest

import jageocoder

import pygeonlp.api as api
from pygeonlp.api.workflow import Workflow
from pygeonlp.api.node import Node


logger = logging.getLogger(__name__)

"""
jageocoder と NEologd を同時に利用する場合のテスト。
環境変数 NEOLOGD_DIC_DIR に NEologd のシステム辞書ディレクトリを
設定しておく必要があります。

このモジュールのテストを行なうには、次のコマンドを実行してください。

python -m unittest -v pygeonlp.tests.test_api_geocoder_neologd
"""

# os.environ["JAGEOCODER_SERVER_URL"] = "https://jageocoder.info-proto.com/jsonrpc"
# os.environ["JAGEOCODER_DB2_DIR"] = "/home/user/jageocoder/db2"
os.environ["NEOLOGD_DIC_DIR"] = os.path.join(
    os.environ.get("HOME"), "neologd/")


class TestModuleMethods(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        # Check 'NEOLOGD_DIC_DIR'
        cls.neologd_dic_dir = os.environ.get('NEOLOGD_DIC_DIR', None)
        if cls.neologd_dic_dir is None:
            return

        # Create test database
        testdir = os.path.abspath(os.path.join(os.getcwd(), 'apitest'))
        os.environ['GEONLP_DB_DIR'] = testdir
        os.makedirs(testdir, 0o755, exist_ok=True)
        dict_manager = api.dict_manager.DictManager(db_dir=testdir)
        dict_manager.setupBasicDatabase()
        api.init()

        # Initialize jageocoder
        try:
            jageocoder.init()
            cls.workflow = Workflow(
                db_dir=testdir,
                system_dic_dir=cls.neologd_dic_dir,
                jageocoder=jageocoder.get_module_tree(),
            )
            cls.parser = cls.workflow.parser
        except jageocoder.exceptions.JageocoderError:
            cls.workflow = None

    def setUp(self):
        if self.neologd_dic_dir is None:
            self.skipTest('環境変数 NEOLOGD_DIC_DIR が設定されていません。')

        if self.workflow is None:
            self.skipTest('住所辞書がインストールされていません。')

    def test_geoparse_with_address(self):
        """
        住所が正しく解析できることを確認する。
        """
        lattice = self.parser.analyze_sentence('千代田区一ツ橋2-1-2')
        lattice_address = self.parser.add_address_candidates(lattice)

        self.assertIsInstance(lattice_address, list)
        self.assertEqual(len(lattice_address), 2)
        self.assertEqual(len(lattice_address[0]), 1)
        self.assertIsInstance(lattice_address[0][0], Node)
        self.assertEqual(lattice_address[0][0].surface, '千代田区一ツ橋2-1')

        # 内部の形態素を確認
        answer = ['千代田区一ツ橋', '2', '-1']
        inner_morphemes = lattice_address[0][0].morphemes
        self.assertEqual(len(inner_morphemes), len(answer))
        for i, node in enumerate(inner_morphemes):
            self.assertIsInstance(node, Node)
            self.assertEqual(node.surface, answer[i])

    def test_geoparse_with_address_no_cutoff(self):
        """
        住所の解析の際に後方の語を途中で切断しないことを確認する。
        """
        lattice = self.parser.analyze_sentence('喜多方市三島町')
        lattice_address = self.parser.add_address_candidates(lattice)
        self.assertIsInstance(lattice_address, list)
        self.assertEqual(len(lattice_address), 2)
        self.assertEqual(lattice_address[0][0].surface, '喜多方市')
        self.assertEqual(lattice_address[1][0].surface, '三島町')

    def test_geoparse_location_disambiguation_hypernym(self):
        """
        住所内の地名語に複数の候補があるとき、
        上位の地名から正しい候補を選択できることを確認する。
        """
        lattice = self.parser.analyze_sentence('横浜市緑区寺山町118番地')
        lattice_address = self.parser.add_address_candidates(lattice)
        self.assertIsInstance(lattice_address, list)
        self.assertEqual(len(lattice_address), 1)
        self.assertEqual(len(lattice_address[0]), 1)
        self.assertIsInstance(lattice_address[0][0], Node)
        self.assertEqual(lattice_address[0][0].surface, '横浜市緑区寺山町118番地')

        # 内部の形態素を確認
        answer = ['横浜市緑区寺山町', '118番', '地']
        inner_morphemes = lattice_address[0][0].morphemes
        self.assertEqual(len(inner_morphemes), len(answer))
        for i, node in enumerate(inner_morphemes):
            self.assertIsInstance(node, Node)
            self.assertEqual(node.surface, answer[i])

    def test_geoparse_location_disambiguation_hyponym(self):
        """
        住所内の地名語に複数の候補があるとき、
        下位の地名から正しい候補を選択できることを確認する。
        """
        lattice = self.parser.analyze_sentence('緑区おゆみ野三丁目15番地3')
        lattice_address = self.parser.add_address_candidates(lattice)
        self.assertIsInstance(lattice_address, list)
        self.assertEqual(len(lattice_address), 2)
        self.assertEqual(len(lattice_address[0]), 1)
        self.assertIsInstance(lattice_address[0][0], Node)
        self.assertEqual(lattice_address[0][0].surface, '緑区おゆみ野三丁目15番地')

        # '緑区おゆみ野' は NEologd で固有名詞に結合されている
        inner_morphemes = lattice_address[0][0].morphemes
        midoriku_oyumino = inner_morphemes[0]
        self.assertEqual('固有名詞', midoriku_oyumino.morphemes['subclass1'])

    def test_geoparse_result(self):
        """
        住所を含む文の解析を geoparse() で実行した結果を確認する。
        """
        answer = [
            {"surface": "NII", "node_type": "NORMAL"},
            {"surface": "は", "node_type": "NORMAL"},
            {"surface": "千代田区一ツ橋2-1", "node_type": "ADDRESS"},
            {"surface": "-2", "node_type": "NORMAL"},
            {"surface": "に", "node_type": "NORMAL"},
            {"surface": "あり", "node_type": "NORMAL"},
            {"surface": "ます", "node_type": "NORMAL"},
            {"surface": "。", "node_type": "NORMAL"},
        ]
        result = self.workflow.geoparse('NIIは千代田区一ツ橋2-1-2にあります。')
        for i, feature in enumerate(result):
            prop = feature['properties']
            self.assertEqual(prop['surface'], answer[i]['surface'])
            self.assertEqual(prop['node_type'], answer[i]['node_type'])

    def test_geoparse_combined_address(self):
        """
        「多摩市落合1-15」の「多摩市落合」が一語として解析されるため
        地名語として抽出できない場合も住所として解析できることを確認する。
        """
        result = self.workflow.geoparse('多摩市落合1-15-2')
        self.assertTrue(
            result[0]['properties']['surface'] in (
                '多摩市落合1-15', '多摩市落合1-15-2'))
        self.assertEqual(result[0]['properties']['node_type'], 'ADDRESS')

    def test_geoparse_combined_address_oneword(self):
        """
        「鹿児島県枕崎市」が一語として解析され、住所表記がそこで完了していても
        住所として解析できることを確認する。
        """
        result = self.workflow.geoparse('鹿児島県枕崎市')
        self.assertEqual(result[0]['properties']['surface'], '鹿児島県枕崎市')
        self.assertEqual(result[0]['properties']['node_type'], 'ADDRESS')

    def test_ma_with_suffix_oneword(self):
        """
        「世田谷区内」のように1語として解析される語に対しても
        suffix（内）を除去した「世田谷区」が抽出できることを確認する。
        """
        nodes = self.parser.service.ma_parseNode('世田谷区内の')
        self.assertEqual(nodes[1]['surface'], '世田谷区')
        self.assertEqual(nodes[1]['subclass2'], '地名語')
        self.assertEqual(nodes[2]['surface'], '内')
        self.assertEqual(nodes[2]['subclass1'], '接尾')

    def test_geoparse_originalform_address(self):
        """
        「東京・世田谷区」の original_form 「東京都世田谷区」を利用して
        住所として解析できることを確認する。
        """
        result = self.workflow.geoparse('東京・世田谷区')
        self.assertEqual(result[0]['properties']['surface'], '東京・世田谷区')
        self.assertEqual(result[0]['properties']['node_type'], 'ADDRESS')
        self.assertEqual(
            result[0]['properties']['address_properties']['fullname'],
            ['東京都', '世田谷区'])



from logging import getLogger
import os
import site
import sys
from typing import Union

from pygeonlp import capi

from pygeonlp.api.dictionary import Dictionary
from pygeonlp.api.metadata import Metadata

logger = getLogger(__name__)


class DictManager(object):
    """
    地名語辞書の管理を行なうクラスです。

    Attributes
    ----------
    _dict_cache : dict
        辞書ID（整数）をキー、辞書identifier（文字列）を値にもつ
        マッピングテーブル（キャッシュ用）。
    capi_ma : pygeonlp.capi オブジェクト
        C 実装の拡張形態素解析機能を利用するためのオブジェクト。
    db_dir: str
        管理対象データベースディレクトリのパス。
    """

    def __init__(self,
                 db_dir: Union[str, bytes, os.PathLike, None] = None):
        """
        このサービスが利用する辞書や、各種設定を初期化します。

        Parameters
        ----------
        db_dir : PathLike, optional
            利用するデータベースディレクトリのパス。
            省略した場合は ``api.get_db_dir()`` で決定します。
        """
        self._dict_cache = {}
        self.capi_ma = None

        if db_dir is None:
            from . import get_db_dir
            db_dir = get_db_dir()

        self.db_dir = db_dir

        # データベースディレクトリの存在チェック
        if os.path.exists(db_dir):
            self.init_capi_ma()

    def init_capi_ma(self):
        # capi_ma オブジェクトを作成
        if self.capi_ma is None:
            capi_options = {'data_dir': str(self.db_dir)}
            self.capi_ma = capi.MA(capi_options)

    def getDictionary(self, id_or_identifier):
        """
        指定した id または identifier を持つ辞書のメタデータを返します。
        id, identifier に一致する辞書が存在しない場合は None を返します。

        Parameters
        ----------
        id_or_identifier : str or int
            str の場合は辞書 identifier で指定。
            int の場合は内部辞書 id で指定。

        Returns
        -------
        Metadata
            Metadata インスタンス。

        Examples
        --------
        >>> from pygeonlp.api.dict_manager import DictManager
        >>> manager = DictManager()
        >>> manager.getDictionary('geonlp:ksj-station-N02')
        {"@context": "https://schema.org/", "@type": "Dataset", "alternateName": "", "creator": [{"@type": "Organization", "name": "株式会社情報試作室", "sameAs": "https://www.info-proto.com/"}], "dateModified": "2021-08-27T17:18:18+09:00", "description": "国土数値情報「鉄道データ（N02）」から作成した、日本の鉄道駅（地下鉄を含む）の辞書です。hypernym には運営者名と路線名を記載しています。「都営」ではなく「東京都」のようになっていますので注意してください。自由フィールドとして、railway_classに「鉄道区分」、institution_typeに「事業者種別」を含みます。", "distribution": [{"@type": "DataDownload", "contentUrl": "https://www.info-proto.com/static/ksj-station-N02.csv", "encodingFormat": "text/csv"}], "identifier": ["geonlp:ksj-station-N02"], "isBasedOn": {"@type": "CreativeWork", "name": "国土数値情報 鉄道データ", "url": "https://nlftp.mlit.go.jp/ksj/gml/datalist/KsjTmplt-N02-v2_2.html"}, "keywords": ["GeoNLP", "地名辞書"], "license": "https://creativecommons.org/licenses/by/4.0/", "name": "国土数値情報 鉄道データ（駅）", "size": "10191", "spatialCoverage": {"@type": "Place", "geo": {"@type": "GeoShape", "box": "26.193265 127.652285 45.41616333333333 145.59723"}}, "temporalCoverage": "../..", "url": "https://www.info-proto.com/static/ksj-station-N02.html"}
        """  # noqa: E501
        self._check_initialized()
        if id_or_identifier is None:
            return self.capi_ma.getDictionaryList()

        jsonld = self.capi_ma.getDictionaryInfo(id_or_identifier)
        if jsonld:
            return Metadata(jsonld)

        return None

    def getDictionaries(self):
        """
        インストール済み辞書のメタデータ一覧を返します。

        Returns
        -------
        list
            Metadata インスタンスのリスト。

        Examples
        --------
        >>> from pygeonlp.api.dict_manager import DictManager
        >>> manager = Manager()
        >>> sorted([x.get_identifier() for x in manager.getDictionaries()])
        ['geonlp:geoshape-city', 'geonlp:geoshape-pref', 'geonlp:ksj-station-N02']
        """
        self._check_initialized()
        dictionaries = []
        for id, metadata in self.capi_ma.getDictionaryList().items():
            # metadata['_internal_id'] = int(id)
            dictionaries.append(Metadata(metadata))

        return dictionaries

    def clearDatabase(self):
        """
        データベースに登録されている辞書をクリアします。
        データベース内の地名語も全て削除されます。

        この関数は、データベースを作り直す際に利用します。

        Examples
        --------
        >>> from pygeonlp.api.dict_manager import DictManager
        >>> manager = DictManager()
        >>> manager.clearDatabase()
        True
        >>> manager.addDictionaryFromWeb(
        ...   'https://geonlp.ex.nii.ac.jp/dictionary/geoshape-city/')
        True
        >>> manager.updateIndex()
        True
        """
        self._check_initialized()
        return self.capi_ma.clearDatabase()

    def addDictionaryFromFile(self, jsonfile, csvfile):
        """
        指定したパスにある辞書メタデータ（JSONファイル）と
        地名解析辞書（CSVファイル）をデータベースに登録します。

        既に同じ identifier を持つ辞書データがデータベースに登録されている場合、
        削除してから新しい辞書データを登録します。

        Parameters
        ----------
        jsonfile : str
            辞書メタデータファイルのパス。
        csvfile : str
            地名解析辞書ファイルのパス。

        Returns
        -------
        bool
            常に True。登録に失敗した場合は例外が発生します。

        Examples
        --------
        >>> from pygeonlp.api.dict_manager import DictManager
        >>> manager = DictManager()
        >>> manager.addDictionaryFromFile(
        ...   'base_data/geoshape-city.json', 'base_data/geoshape-city.csv')
        True
        >>> manager.updateIndex()
        True
        """
        self._check_initialized()
        if not isinstance(jsonfile, str) or not isinstance(csvfile, str):
            raise TypeError("jsonfile と csvfile は str で指定してください。")

        dic = Dictionary.load(jsonfile, csvfile)
        new_identifier = dic.get_identifier()
        if self.getDictionary(new_identifier):
            self.removeDictionary(new_identifier)

        return dic.add(self.capi_ma)

    def addDictionaryFromWeb(self, url, params=None, **kwargs):
        """
        指定した URL にあるページに含まれる辞書メタデータ（JSON-LD）を取得し、
        メタデータに記載されている URL から地名解析辞書（CSVファイル）を取得し、
        データベースに登録します。

        既に同じ identifier を持つ辞書データがデータベースに登録されている場合、
        削除してから新しい辞書データを登録します。

        Parameters
        ----------
        url : str
            辞書メタデータを含むウェブページの URL。
        params : dict, optional
            requests.get に渡す params パラメータ。
        **kwargs : dict, optional
            requests.get に渡す kwargs パラメータ。

        Returns
        -------
        bool
            常に True。登録に失敗した場合は例外が発生します。

        Examples
        --------
        >>> from pygeonlp.api.dict_manager import DictManager
        >>> manager = DictManager()
        >>> manager.clearDatabase()
        True
        >>> manager.addDictionaryFromWeb('https://geonlp.ex.nii.ac.jp/dictionary/geoshape-city/')
        True
        >>> manager.updateIndex()
        True
        """
        self._check_initialized()
        dic = Dictionary.download(url, params, **kwargs)
        new_identifier = dic.get_identifier()
        if self.getDictionary(new_identifier):
            self.removeDictionary(new_identifier)

        return dic.add(self.capi_ma)

    def addDictionaryFromCsv(self, csvfile, name=None, identifier=None):
        """
        指定したパスにある地名解析辞書（CSVファイル）をデータベースに登録します。

        辞書の name と identifier を省略した場合、
        CSV ファイル名から自動的に設定されます。

        既に同じ identifier を持つ辞書データがデータベースに登録されている場合、
        削除してから新しい辞書データを登録します。

        Parameters
        ----------
        csvfile : str
            地名解析辞書ファイルのパス。
        name : str, optional
            辞書名。省略した場合、 CSV ファイルの basename を利用します。
        identifier : str, optional
            辞書 identifier。省略した場合、 CSV ファイルの basename を取り、
            ``geonlp:<basename>`` を利用します。

        Returns
        -------
        bool
            常に True。登録に失敗した場合は例外が発生します。

        Examples
        --------
        >>> from pygeonlp.api.dict_manager import DictManager
        >>> manager = DictManager()
        >>> manager.addDictionaryFromCsv(
        ...   'base_data/ksj-station-N02-2020.csv',
        ...   name='日本の鉄道駅（2020年）')
        True
        >>> manager.updateIndex()
        True
        """
        self._check_initialized()
        if not isinstance(csvfile, str):
            raise TypeError("csvfile は str で指定してください。")

        dic = Dictionary.create(csvfile, name, identifier)
        new_identifier = dic.get_identifier()
        if self.getDictionary(new_identifier):
            self.removeDictionary(new_identifier)

        return dic.add(self.capi_ma)

    def saveDictionaryFromWeb(self, jsonfile, csvfile, url,
                              params=None, **kwargs):
        """
        指定した URL にあるページに含まれる辞書メタデータ（JSON-LD）を取得し、
        メタデータに記載されている URL から地名解析辞書（CSVファイル）を取得し、
        指定されたファイルに保存します。

        Parameters
        ----------
        jsonfile : str
            json-ld を保存するファイル名。
        csvfile : str
            CSV データを保存するファイル名。
        url : str
            辞書メタデータを含むウェブページの URL。
        params : dict, optional
            requests.get に渡す params パラメータ。
        **kwargs : dict, optional
            requests.get に渡す kwargs パラメータ。

        Returns
        -------
        bool
            常に True。ダウンロードに失敗した場合は例外が発生します。

        Examples
        --------
        >>> from pygeonlp.api.dict_manager import DictManager
        >>> manager = DictManager()
        >>> manager.saveDictionaryFromWeb('geoshape.json', 'geoshape.csv', 'https://geonlp.ex.nii.ac.jp/dictionary/geoshape-city/')
        True
        >>> import os
        >>> os.remove('geoshape.json')
        >>> os.remove('geoshape.csv')
        """  # noqa: E501
        self._check_initialized()
        dic = Dictionary.download(url, params, **kwargs)
        return dic.save(jsonfile, csvfile)

    def removeDictionary(self, identifier):
        """
        identifier で指定した辞書をデータベースから削除します。

        Parameters
        ----------
        identifier : str
            辞書の identifier ("geonlp:xxxxx")。

        Returns
        -------
        bool
            常に True。

        Raises
        ------
        RuntimeError
            指定した辞書が登録されていない場合は RUntimeError が発生します。

        Examples
        --------
        >>> from pygeonlp.api.dict_manager import DictManager
        >>> manager = DictManager()
        >>> manager.removeDictionary('geonlp:geoshape-city')
        True
        >>> manager.updateIndex()
        True
        """
        self._check_initialized()
        ret = self.capi_ma.removeDictionary(identifier)
        return ret

    def updateIndex(self):
        """
        辞書のインデックスを更新して検索可能にします。

        Examples
        --------
        >>> from pygeonlp.api.dict_manager import DictManager
        >>> manager = DictManager()
        >>> manager.clearDatabase()
        True
        >>> manager.addDictionaryFromWeb('https://geonlp.ex.nii.ac.jp/dictionary/geoshape-city/')
        True
        >>> manager.updateIndex()
        True
        """
        self._check_initialized()
        return self.capi_ma.updateIndex()

    @staticmethod
    def get_package_files():
        """
        pip list コマンドを実行し、pygeonlp パッケージとしてインストールされた
        ファイルのフルパス一覧を取得します。

        参考： https://pip.pypa.io/en/latest/user_guide/#using-pip-from-your-program
        """
        import subprocess
        import pygeonlp.api

        show_args = [sys.executable, '-m',
                     'pip', 'show', '--files', 'pygeonlp']
        result = subprocess.check_output(show_args).decode('utf-8')
        lines = result.split("\n")
        files = None
        for i, line in enumerate(lines):
            if line.strip() == "Files:":
                files = lines[i+1:]
                break

        if files is None:
            raise RuntimeError("パッケージ内のファイル一覧が取得できません。")

        # pygeonlp.api.__file__ は '../site-packages/pygeonlp/api/__init__.py'
        # '../site-packages/' を base_path としてフルパスを取得する
        base_path = os.path.abspath(
            os.path.join(pygeonlp.api.__file__, '../../..'))
        files = [os.path.abspath(os.path.join(base_path, x.strip()))
                 for x in files]
        return files

    def setupBasicDatabase(self, src_dir=None):
        """
        基本的な地名解析辞書を登録したデータベースを作成します。

        Parameters
        ----------
        src_dir : str, optional
            地名解析辞書ファイルが配置されているディレクトリ。
            省略された場合には、 ``sys.prefix`` または ``site.USER_BASE`` の
            下に ``pygeonlp_basedata`` がないか探します。
            見つからない場合は ``RuntimeError`` を送出しますので、
            ディレクトリを指定してください。
        """
        data_dir = None
        base_dir = os.getcwd()
        candidates = []
        if src_dir:
            candidates.append(src_dir)

        candidates += [
            os.path.join(sys.prefix, 'pygeonlp_basedata'),
            os.path.join(site.USER_BASE, 'pygeonlp_basedata'),
            os.path.join(base_dir, 'base_data'),
            os.path.join(base_dir, 'pygeonlp_basedata'),
        ]

        for cand in candidates:
            if os.path.exists(cand):
                data_dir = cand
                break

        if not data_dir:
            files = self.__class__.get_package_files()
            for path in files:
                pos = path.find('/pygeonlp_basedata')
                if pos < 0:
                    continue

                data_dir = path[0:pos + len('/pygeonlp_basedata')]

        if not data_dir:
            raise RuntimeError("地名解析辞書がインストールされたディレクトリが見つかりません。")

        os.makedirs(self.db_dir, 0o755, exist_ok=True)
        self.init_capi_ma()

        updated = False
        if self.getDictionary('geonlp:geoshape-city') is None:
            self.addDictionaryFromFile(
                jsonfile=os.path.join(data_dir, 'geoshape-city.json'),
                csvfile=os.path.join(data_dir, 'geoshape-city.csv'))
            updated = True

        if self.getDictionary('geonlp:geoshape-pref') is None:
            self.addDictionaryFromFile(
                jsonfile=os.path.join(data_dir, 'geoshape-pref.json'),
                csvfile=os.path.join(data_dir, 'geoshape-pref.csv'))
            updated = True

        if self.getDictionary('geonlp:ksj-station-N02') is None:
            self.addDictionaryFromFile(
                jsonfile=os.path.join(data_dir, 'ksj-station-N02.json'),
                csvfile=os.path.join(data_dir, 'ksj-station-N02.csv'))
            updated = True

        if updated:
            self.updateIndex()

    def _check_initialized(self):
        """
        capi オブジェクトが初期化されていることを確認します。
        """
        if self.capi_ma is None:
            raise RuntimeError("CAPI が初期化されていません。")

    def _add_dict_identifier(self, word):
        """
        語に含まれる辞書の内部 ID "dictionary_id" を見て
        語に辞書の識別子 "dictionary_identifier" を追加します。
        """
        if word is None or 'dictionary_id' not in word or \
                'dictionary_identifier' in word:
            return word

        dictionary_id = word['dictionary_id']
        if dictionary_id in self._dict_cache:
            identifier = self._dict_cache[dictionary_id]
        else:
            identifier = self.capi_ma.getDictionaryIdentifierById(
                dictionary_id)

            self._dict_cache[dictionary_id] = identifier

        word['dictionary_identifier'] = identifier
        return word



"""
scoring モジュールは、一通りのジオパージング処理を行なうために必要最小限の、
スコアリング処理の実装をまとめています。
十分な精度が得られない場合、用途に応じてチューニングを行なうか、
別の手法を実装して利用してください。
"""

from logging import getLogger

from pygeonlp.api.linker import Evaluator
from pygeonlp.api.node import Node

logger = getLogger(__name__)


class ScoringClass(object):

    def __init__(self, options=None):
        """
        スコアリングクラスの初期化を行ないます。

        独自のスコアリング手法を実装したクラスを利用して
        アノテーション結果をランキングしたい場合、
        このクラスから派生したクラスを定義し、
        pygeonlp.api.geoparse() を呼びだす際に scoring_class パラメータに
        利用するクラス名を指定してください。

        Parameters
        ----------
        options : any
            パスのスコアリングやノード間のスコアを計算するときに
            利用するパラメータです。

            geoparse() の ``scoring_options`` で指定した値が入ります。
            デフォルト実装では int の値をとり、
            地名語間の関係や距離を計算する際に、いくつ先の形態素まで
            照合するかを指定します。デフォルトは5です。
            地名語や住所の候補を含まない形態素はカウントしません。

            独自のクラスを派生する場合には任意の型が利用できます。
        """
        self.options = options

    def path_score(self, path):
        """
        パス表現の結果を一つ受け取り、スコアを計算して返すデフォルト実装です。

        Parameters
        ----------
        path : list of Node
            解析結果候補のパス表現。

        Returns
        -------
        int
            与えられたパスを評価したスコアを返します。
            値が大きいパスほど geoparse() の結果で上位候補として返されます。

        Note
        ----
        次のように得点を計算しています。

        - 地名語・住所を候補として全く含まない形態素が存在する場合

            +形態素数 ×1点（地名語か非地名語か不明な形態素は非地名語を優先）

        - 全ての候補の中に住所ノード候補が含まれている場合

            +住所階層数（'東京都/千代田区'なら2）×10点

        - 全ての候補の中に含まれる地名語ノード候補に、
          同じ固有名クラスが2回以上出現する場合

            +出現回数 ×10点

        - ある形態素に地名語ノード候補が含まれる場合

            nlookup 先までの形態素に含まれる地名語・住所ノード候補との
            関係によるスコアを node_relation_score() で計算し加算
        """
        if self.options:
            nlookup = self.options
        else:
            nlookup = 5

        if not isinstance(nlookup, int):
            raise RuntimeError(
                "オプションパラメータは整数値で指定してください。")

        score = 0
        geowords = Evaluator.collect_geowords(path)
        addresses = Evaluator.collect_addresses(path)
        geonodes = geowords + addresses
        score += len(path) - len(geonodes)  # 住所・地名語以外は +1

        # 住所による得点を計算
        for address in addresses:
            score += 10 * len(address.morphemes)  # 住所階層数 * 10

        # 同じ固有名クラスが複数含まれていると得点
        ne_classes = {}
        for geoword in geowords:
            ne_class = geoword.prop['ne_class']
            if ne_class in ne_classes:
                ne_classes[ne_class] += 1
            else:
                ne_classes[ne_class] = 1

        for ne_class, freq in ne_classes.items():
            if freq > 1:
                logger.debug("固有名クラス {} が {} 件存在 (*10)".format(
                    ne_class, freq))
                score += freq * 10

        # ノード間の距離を計算
        # nlookup ノード先まで確認する
        for i0 in range(len(path)):
            n0 = path[i0]
            if n0.node_type == Node.NORMAL:
                continue

            for i1 in range(i0 + 1, len(path)):
                n1 = path[i1]
                if n1.node_type == Node.NORMAL:
                    continue

                score += self.node_relation_score(n0, n1)
                nlookup -= 1
                if nlookup <= 0:
                    break

        return score

    def node_relation_score(self, node0, node1):
        """
        node クラスのインスタンス node0 と node1 との関係による
        スコアを計算するデフォルト実装です。
        関係が強いノードほど高いスコアを返します。

        Examples
        --------
        >>> import pygeonlp.api as api
        >>> from pygeonlp.api.scoring import ScoringClass
        >>> api.init()
        >>> s = ScoringClass()
        >>> s.node_relation_score(api.analyze('国会議事堂前')[0][0],
        ...   api.analyze('永田町')[0][0])
        20

        Note
        ----
        スコアは以下のように計算しています。

        - どちらかのノードが地名語でも住所でもない（NORMAL）場合
            0 点
        - どちらのノードも地名語で、固有名クラスが一致する場合（同クラス）
            +10 点
        - どちらのノードがもう一方の hypernym にふくまれる場合（親子関係）
            +5 点
        - 両方のノードの hypernym が1つ以上一致する場合（兄弟関係）
            +5 点
        - ノード間の距離（単位は degree）を計算し、
            0.2度以下の場合は +5 点、それ以上離れている場合は +1/距離を加算
        """
        if node0.node_type == Node.NORMAL or node1.node_type == Node.NORMAL:
            return 0

        score = 0

        if node0.node_type == Node.GEOWORD and node1.node_type == Node.GEOWORD:
            # 固有名クラスの一致度
            if node0.prop['ne_class'] == node1.prop['ne_class']:
                score += 10

        # 上位下位関係
        hypernym_node0 = None
        hypernym_node1 = None
        if node0.node_type == Node.GEOWORD:
            hypernym_node0 = set(node0.prop.get('hypernym', []))
            if hypernym_node0 & node1.get_notations():
                score += 5

        if node1.node_type == Node.GEOWORD:
            hypernym_node1 = set(node1.prop.get('hypernym', []))
            if hypernym_node1 & node0.get_notations():
                score += 5

        if hypernym_node0 and hypernym_node1 and \
           hypernym_node0 & hypernym_node1:
            # 兄弟関係
            score += 5

        # 相互距離
        try:
            dist = node0.distance(node1)
            if dist < 10000.0:  # 10km 以下
                score += 5
            else:
                # 50km までは 1 点追加
                score += int(50000.0 / dist)
        except RuntimeError:
            pass

        logger.debug("{} と {} のスコア => {}".format(
            node0.simple(), node1.simple(), score))

        return score



import logging
import os
import unittest

import pygeonlp.api as api

logger = logging.getLogger(__name__)

"""
doc_string に書くには複雑過ぎる API のテスト。

このモジュールのテストを行なうには、次のコマンドを実行してください。

python -m unittest -v pygeonlp.tests.test_api
"""


class TestModuleMethods(unittest.TestCase):

    service = api.service.Service()

    def setUp(self):
        testdir = os.path.abspath(os.path.join(os.getcwd(), 'apitest'))
        os.makedirs(testdir, 0o755, exist_ok=True)
        self.dictmanager = api.dict_manager.DictManager(db_dir=testdir)
        self.dictmanager.setupBasicDatabase()
        api.init(db_dir=testdir)

    def test_search_word(self):
        words = self.service.searchWord('神保町')
        self.assertIsInstance(words, dict)
        self.assertIn('uN6ecI', words)  # 新宿線神保町駅

    def test_set_dictionaries(self):
        # Set active dictionaries and check the results
        self.service.setActiveDictionaries(pattern=r'.*')
        words = self.service.searchWord('和歌山市')
        self.assertIsInstance(words, dict)
        self.assertEqual(len(words.keys()), 4)

        # Disactivate all dictionaries
        self.service.disactivateDictionaries(pattern=r'.*')
        words = self.service.searchWord('和歌山市')
        self.assertIsInstance(words, dict)
        self.assertEqual(len(words.keys()), 0)

    def test_set_active_classes(self):
        # Set active classes and check the results
        classes = [".*", "-鉄道施設/.*"]
        self.service.setActiveClasses(classes)
        words = self.service.searchWord('神保町')
        self.assertIsInstance(words, dict)
        self.assertNotIn('AGGwyc', words)  # 新宿線神保町駅は鉄道施設

        # "鉄道施設/.*" は除外するが "駅$" は除外しない
        self.service.setActiveClasses([".*", "-鉄道施設/.*", r'.*駅$'])
        words = self.service.searchWord('神保町')
        self.assertIsInstance(words, dict)
        self.assertIn('AGGwyc', words)  # 新宿線神保町駅は鉄道駅

        # Reset active classes
        api.setActiveClasses()
        words = self.service.searchWord('神保町')
        self.assertIsInstance(words, dict)
        self.assertIn('AGGwyc', words)  # 新宿線神保町駅

    def test_geo_contains_filter(self):
        from pygeonlp.api.spatial_filter import SpatialFilter, GeoContainsFilter
        geojson = SpatialFilter.get_geometry((
            'https://geoshape.ex.nii.ac.jp/city/geojson/'
            '20230101/13/13206A1968.geojson')).ExportToJson()
        gcfilter = GeoContainsFilter(geojson)
        api.default_workflow().filters = [gcfilter]
        results = api.geoparse('府中に行きます')
        self.assertEqual(results[0]['properties']['node_type'], 'GEOWORD')
        api.default_workflow().filters = []

    def test_geo_disjoint_filter(self):
        from pygeonlp.api.spatial_filter import SpatialFilter, GeoDisjointFilter
        geojson = SpatialFilter.get_geometry((
            'https://geoshape.ex.nii.ac.jp/city/geojson/'
            '20200101/13/13208A1968.geojson')).ExportToJson()
        gcfilter = GeoDisjointFilter(geojson)
        api.default_workflow().filters = [gcfilter]
        results = api.geoparse('府中に行きます')
        self.assertEqual(results[0]['properties']['node_type'], 'GEOWORD')
        api.default_workflow().filters = []

    def test_issue1(self):
        """
        RecursionError: maximum recursion depth exceeded
        """
        results = api.geoparse("第二希望：東京、静岡、三重、滋賀、大阪")
        self.assertEqual(len(results), 13)
        self.assertEqual(results[4]['properties']['node_type'], 'GEOWORD')

    def test_issue15(self):
        """
        「〇〇線××駅」を正しく解析できない
        """
        results = api.geoparse("高崎線上尾駅は埼玉県にあります。")
        self.assertEqual(results[0]['properties']['surface'], "高崎")
        self.assertEqual(results[1]['properties']['surface'], "線")
        self.assertEqual(results[2]['properties']['surface'], "上尾駅")


if __name__ == '__main__':
    unittest.main()



"""
スコアリング方法のカスタマイズ
サンプルコード

このコードをテストするには以下のコマンドを実行してください。
python sample_myscore.py
"""
import pygeonlp.api as api
from pygeonlp.api.linker import Evaluator
from pygeonlp.api.scoring import ScoringClass

api.init()


class MyScoringClass(ScoringClass):

    def path_score(self, path):
        """
        パスの中に指定した文字列で始まる固有名クラスの地名語が
        存在する数をスコアとして返すスコアリングメソッド。

        Parameters
        ----------
        path : list of Node
            解析結果候補のパス表現。
        self.options : str
            カウントする固有名クラスの先頭文字列

        Returns
        -------
        int
            target_class にマッチする固有名クラスを持つ地名語数。
        """
        if not isinstance(self.options, str):
            raise RuntimeError(
                "オプションパラメータは文字列で指定してください。")

        target_class = self.options
        score = 0
        geowords = Evaluator.collect_geowords(path)
        for geoword in geowords:
            if geoword.prop['ne_class'].startswith(target_class):
                score += 1

        return score


if __name__ == '__main__':
    print("'鉄道施設' が多い候補を優先した場合。")
    api.init(scoring_class=MyScoringClass, scoring_options='鉄道施設')
    print(api.geoparse('和歌山市は晴れ。'))
    print("'市区町村' が多い候補を優先した場合。")
    api.init(scoring_class=MyScoringClass, scoring_options='市区町村')
    print(api.geoparse('和歌山市は晴れ。'))



from logging import getLogger
import os
import warnings

from deprecated import deprecated
import jageocoder

from pygeonlp.api.dict_manager import DictManager
from pygeonlp.api.workflow import Workflow, WorkflowError

logger = getLogger(__name__)
_default_workflow = None
_default_manager = None

__version__ = '1.2.4'


def get_db_dir():
    """
    データベースディレクトリを取得します。
    データベースディレクトリは次の優先順位に従って決定します。

    - 環境変数 ``GEONLP_DB_DIR`` の値
    - 環境変数 ``HOME`` が指すディレクトリの下の ``geonlp/db/``

    どちらの環境変数も指定されていない場合は ``RuntimeError`` を送出して
    終了します。

    Return
    ------
    str
        ディレクトリの絶対パス。
    """
    # 環境変数 GEONLP_DB_DIR をチェック
    db_dir = os.environ.get('GEONLP_DB_DIR')
    if db_dir:
        db_dir = os.path.abspath(db_dir)

    # 環境変数 HOME が利用できれば $HOME/geonlp/db
    home = os.environ.get('HOME')
    if not db_dir and home:
        db_dir = os.path.join(home, 'geonlp/db')
        db_dir = os.path.abspath(db_dir)

    if not db_dir:
        raise RuntimeError(("データベースディレクトリを環境変数 "
                            "GEONLP_DB_DIR で指定してください。"))

    return db_dir


def get_jageocoder_db_dir():
    """
    **deprecated**

    jageocoder の辞書が配置されているディレクトリを取得します。
    次の優先順位に従って決定します。

    - 環境変数 ``JAGEOCODER_DB_DIR`` の値
    - 環境変数 ``HOME`` が指すディレクトリの下の ``jageocoder/db``
    - どちらの環境変数も指定されていない場合は None

    Return
    ------
    str
        ディレクトリの絶対パス、または None。
    """
    warnings.warn(('get_jageocoder_db_dir() はver.1.2で廃止予定です。'
                   'jageocoder.get_db_dir() を利用してください。'),
                  DeprecationWarning)

    return jageocoder.get_db_dir()


def init(db_dir=None, **options):
    """
    API をデフォルトパラメータで初期化します。

    Parameters
    ----------
    db_dir : PathLike, optional
        データベースディレクトリ。
        省略した場合は ``api.get_db_dir()`` が返す値を利用します。

    Examples
    --------
    >>> import pygeonlp.api as api
    >>> api.init()

    Note
    ----
    実際には、この関数は初期化した Workflow オブジェクト ``_default_workflow``
    および DictManager オブジェクト ``_default_manager`` を作成し、
    利用可能な状態にします。
    ``pygeonlp.api`` の各関数は、これらのオブジェクトの
    メンバ関数を呼びだすヘルパー関数として実装されています。

    Workflow のパラメータを変更したい場合、 ``api.default_workflow()`` で
    Workflow オブジェクトを取得して ``set_parser()`` などを呼びだします。

    異なるパラメータで初期化した Workflow オブジェクトを生成し、
    直接そのメンバ関数を呼びだすことも可能です。
    """  # noqa: E501
    global _default_workflow, _default_manager
    if _default_workflow:
        del _default_workflow

    if _default_manager:
        del _default_manager

    _default_workflow = Workflow(db_dir=db_dir, **options)
    _default_manager = DictManager(db_dir=db_dir)


@deprecated(
    version='1.2.0',
    reason=(
        'This is a low level API. Call class methods of'
        ' "pygeonlp.api.service.Service" directly.')
)
def ma_parse(sentence):
    """
    センテンスを形態素解析した結果を MeCab 互換の文字列として返します。
    *deprecated* この関数は低水準 API なので、廃止予定です。

    Parameters
    ----------
    sentence : str
        解析する文字列。

    Returns
    -------
    str
        解析結果の改行区切りテキスト。

    Examples
    --------
    >>> import pygeonlp.api as api
    >>> api.init()
    >>> print(api.ma_parse('今日は国会議事堂前まで歩きました。'))
    <BLANKLINE>
    今日    名詞,副詞可能,*,*,*,*,今日,キョウ,キョー
    は      助詞,係助詞,*,*,*,*,は,ハ,ワ
    国会議事堂前    名詞,固有名詞,地名語,Bn4q6d:国会議事堂前駅/cE8W4w:国会議事堂前駅,*,*,国会議事堂前,,
    まで    助詞,副助詞,*,*,*,*,まで,マデ,マデ
    歩き    動詞,自立,*,*,五段・カ行イ音便,連用形,歩く,アルキ,アルキ
    まし    助動詞,*,*,*,特殊・マス,連用形,ます,マシ,マシ
    た      助動詞,*,*,*,特殊・タ,基本形,た,タ,タ
    。      記号,句点,*,*,*,*,。,。,。
    EOS
    <BLANKLINE>
    """  # noqa: E501
    _check_initialized()
    return _default_workflow.parser.service.ma_parse(sentence)


@deprecated(
    version='1.2.0',
    reason=(
        'This is a low level API. Call class methods of '
        '"pygeonlp.api.service.Service" directly.'
    )
)
def ma_parseNode(sentence):
    """
    センテンスを形態素解析した結果を MeCab 互換のノード配列として返します。
    *deprecated* この関数は低水準 API なので、廃止予定です。

    Parameters
    ----------
    sentence : str
        解析する文字列。

    Returns
    -------
    list
        解析結果のリスト。

    Example
    -------
    >>> import pygeonlp.api as api
    >>> api.init()
    >>> api.ma_parseNode('今日は国会議事堂前まで歩きました。')
    [{'conjugated_form': '*', 'conjugation_type': '*', 'original_form': '*', 'pos': 'BOS/EOS', 'prononciation': '*', 'subclass1': '*', 'subclass2': '*', 'subclass3': '*', 'surface': '', 'yomi': '*'}, {'conjugated_form': '*', 'conjugation_type': '*', 'original_form': '今日', 'pos': '名詞', 'prononciation': 'キョー', 'subclass1': '副詞可能', 'subclass2': '*', 'subclass3': '*', 'surface': '今日', 'yomi': 'キョウ'}, {'conjugated_form': '*', 'conjugation_type': '*', 'original_form': 'は', 'pos': '助詞', 'prononciation': 'ワ', 'subclass1': '係助詞', 'subclass2': '*', 'subclass3': '*', 'surface': 'は', 'yomi': 'ハ'}, {'conjugated_form': '*', 'conjugation_type': '*', 'original_form': '国会議事堂前', 'pos': '名詞', 'prononciation': '', 'subclass1': '固有名詞', 'subclass2': '地名語', 'subclass3': 'Bn4q6d:国会議事堂前駅/cE8W4w:国会議事堂前駅', 'surface': '国会議事堂前', 'yomi': ''}, {'conjugated_form': '*', 'conjugation_type': '*', 'original_form': 'まで', 'pos': '助詞', 'prononciation': 'マデ', 'subclass1': '副助詞', 'subclass2': '*', 'subclass3': '*', 'surface': 'まで', 'yomi': 'マデ'}, {'conjugated_form': '五段・カ行イ音便', 'conjugation_type': '連用形', 'original_form': '歩く', 'pos': '動詞', 'prononciation': 'アルキ', 'subclass1': '自立', 'subclass2': '*', 'subclass3': '*', 'surface': '歩き', 'yomi': 'アルキ'}, {'conjugated_form': '特殊・マス', 'conjugation_type': '連用形', 'original_form': 'ます', 'pos': '助動詞', 'prononciation': 'マシ', 'subclass1': '*', 'subclass2': '*', 'subclass3': '*', 'surface': 'まし', 'yomi': 'マシ'}, {'conjugated_form': '特殊・タ', 'conjugation_type': '基本形', 'original_form': 'た', 'pos': '助動詞', 'prononciation': 'タ', 'subclass1': '*', 'subclass2': '*', 'subclass3': '*', 'surface': 'た', 'yomi': 'タ'}, {'conjugated_form': '*', 'conjugation_type': '*', 'original_form': '。', 'pos': '記号', 'prononciation': '。', 'subclass1': '句点', 'subclass2': '*', 'subclass3': '*', 'surface': '。', 'yomi': '。'}, {'conjugated_form': '*', 'conjugation_type': '*', 'original_form': '*', 'pos': 'BOS/EOS', 'prononciation': '*', 'subclass1': '*', 'subclass2': '*', 'subclass3': '*', 'surface': '', 'yomi': '*'}]

    """  # noqa: E501
    _check_initialized()
    return _default_workflow.parser.service.ma_parseNode(sentence)


@deprecated(
    version='1.2.0',
    reason=(
        'This is a low level API. Call class methods of '
        '"pygeonlp.api.service.Service" directly.'
    )
)
def getWordInfo(geolod_id):
    """
    指定した geolod_id を持つ語の情報を返します。
    id が辞書に存在しない場合は None を返します。
    *deprecated* この関数は低水準 API なので、廃止予定です。

    Parameters
    ----------
    geolod_id : str or Iterable
        語の ID または ID のリストを返すイテレータ。

    Returns
    -------
    dict
        geolod_id をキー、語の情報を値に持つ dict。

    Examples
    --------
    >>> import pygeonlp.api as api
    >>> api.init()
    >>> api.getWordInfo('Bn4q6d')
    {'body': '国会議事堂前', 'dictionary_id': 3, 'entry_id': 'LrGGxY', 'geolod_id': 'Bn4q6d', 'hypernym': ['東京地下鉄', '4号線丸ノ内線'], 'institution_type': '民営鉄道', 'latitude': '35.674845', 'longitude': '139.74534166666666', 'ne_class': '鉄道施設/鉄道駅', 'railway_class': '普通鉄道', 'suffix': ['駅', ''], 'dictionary_identifier': 'geonlp:ksj-station-N02'}
    """  # noqa: E501
    _check_initialized()
    return _default_workflow.parser.service.getWordInfo(geolod_id)


@deprecated(
    version='1.2.0',
    reason=(
        'This is a low level API. Call class methods of '
        '"pygeonlp.api.service.Service" directly.'
    )
)
def searchWord(key):
    """
    指定した表記または読みを持つ語の情報を返します。
    一致する語が辞書に存在しない場合は None を返します。
    *deprecated* この関数は低水準 API なので、廃止予定です。

    Parameters
    ----------
    key : str
        語の表記または読み。

    Returns
    -------
    dict
        geolod_id をキー、語の情報を値に持つ dict。

    Examples
    --------
    >>> import pygeonlp.api as api
    >>> api.init()
    >>> api.searchWord('国会議事堂前')
    {'Bn4q6d': {'body': '国会議事堂前', 'dictionary_id': 3, 'entry_id': 'LrGGxY', 'geolod_id': 'Bn4q6d', 'hypernym': ['東京地下鉄', '4号線丸ノ内線'], 'institution_type': '民営鉄道', 'latitude': '35.674845', 'longitude': '139.74534166666666', 'ne_class': '鉄道施設/鉄道駅', 'railway_class': '普通鉄道', 'suffix': ['駅', ''], 'dictionary_identifier': 'geonlp:ksj-station-N02'}, 'cE8W4w': {'body': '国会議事堂前', 'dictionary_id': 3, 'entry_id': '4NFELa', 'geolod_id': 'cE8W4w', 'hypernym': ['東京地下鉄', '9号線千代田線'], 'institution_type': '民営鉄道', 'latitude': '35.673543333333335', 'longitude': '139.74305333333334', 'ne_class': '鉄道施設/鉄道駅', 'railway_class': '普通鉄道', 'suffix': ['駅', ''], 'dictionary_identifier': 'geonlp:ksj-station-N02'}}
    """  # noqa: E501
    _check_initialized()
    return _default_workflow.parser.service.searchWord(key)


@deprecated(
    version='1.2.0',
    reason=(
        'This is a low level API. Call class methods of '
        '"pygeonlp.api.service.Service" directly.'
    )
)
def getDictionary(id_or_identifier):
    """
    指定した id または identifier を持つ辞書のメタデータを返します。
    id, identifier に一致する辞書が存在しない場合は None を返します。

    Parameters
    ----------
    id_or_identifier : str or int
        str の場合は辞書 identifier で指定。
        int の場合は内部辞書 id で指定。

    Returns
    -------
    Metadata
        Metadata インスタンス。

    Examples
    --------
    >>> import pygeonlp.api as api
    >>> api.init()
    >>> api.getDictionary('geonlp:ksj-station-N02')
    {"@context": "https://schema.org/", "@type": "Dataset", "alternateName": "", "creator": [{"@type": "Organization", "name": "株式会社情報試作室", "sameAs": "https://www.info-proto.com/"}], "dateModified": "2021-08-27T17:18:18+09:00", "description": "国土数値情報「鉄道データ（N02）」から作成した、日本の鉄道駅（地下鉄を含む）の辞書です。hypernym には運営者名と路線名を記載しています。「都営」ではなく「東京都」のようになっていますので注意してください。自由フィールドとして、railway_classに「鉄道区分」、institution_typeに「事業者種別」を含みます。", "distribution": [{"@type": "DataDownload", "contentUrl": "https://www.info-proto.com/static/ksj-station-N02.csv", "encodingFormat": "text/csv"}], "identifier": ["geonlp:ksj-station-N02"], "isBasedOn": {"@type": "CreativeWork", "name": "国土数値情報 鉄道データ", "url": "https://nlftp.mlit.go.jp/ksj/gml/datalist/KsjTmplt-N02-v2_2.html"}, "keywords": ["GeoNLP", "地名辞書"], "license": "https://creativecommons.org/licenses/by/4.0/", "name": "国土数値情報 鉄道データ（駅）", "size": "10191", "spatialCoverage": {"@type": "Place", "geo": {"@type": "GeoShape", "box": "26.193265 127.652285 45.41616333333333 145.59723"}}, "temporalCoverage": "../..", "url": "https://www.info-proto.com/static/ksj-station-N02.html"}
    """  # noqa: E501
    _check_initialized()
    return _default_manager.getDictionary(id_or_identifier)


@deprecated(
    version='1.2.0',
    reason=(
        'This is a low level API. Call class methods of '
        '"pygeonlp.api.dict_manager.DictManager" directly.'
    )
)
def getDictionaries():
    """
    インストール済み辞書のメタデータ一覧を返します。

    Returns
    -------
    list
        Metadata インスタンスのリスト。

    Examples
    --------
    >>> import pygeonlp.api as api
    >>> api.init()
    >>> sorted([x.get_identifier() for x in api.getDictionaries()])
    ['geonlp:geoshape-city', 'geonlp:geoshape-pref', 'geonlp:ksj-station-N02']
    """
    _check_initialized()
    return _default_manager.getDictionaries()


@deprecated(
    version='1.2.0',
    reason=(
        'This is a low level API. Call class methods of '
        '"pygeonlp.api.dict_manager.DictManager" directly.'
    )
)
def clearDatabase():
    """
    地名語データベースに登録されている辞書をクリアします。
    データベース内の地名語も全て削除されます。

    この関数は、データベースを作り直す際に利用します。
    """
    _check_initialized()
    _default_manager.clearDatabase()


@deprecated(
    version='1.2.0',
    reason=(
        'This is a low level API. Call class methods of '
        '"pygeonlp.api.dict_manager.DictManager" directly.'
    )
)
def addDictionaryFromFile(jsonfile, csvfile):
    """
    指定したパスにある辞書メタデータ（JSONファイル）と
    地名解析辞書（CSVファイル）をデータベースに登録します。

    既に同じ identifier を持つ辞書データがデータベースに登録されている場合、
    削除してから新しい辞書データを登録します。

    登録した辞書を利用可能にするには、 ``setActivateDictionaries()``
    または ``activateDictionaires()`` で有効化する必要があります。

    Parameters
    ----------
    jsonfile : str
        辞書メタデータファイルのパス。
    csvfile : str
        地名解析辞書ファイルのパス。

    Returns
    -------
    bool
        常に True。登録に失敗した場合は例外が発生します。
    """  # noqa: E501
    _check_initialized()
    return _default_manager.addDictionaryFromFile(jsonfile, csvfile)


@deprecated(
    version='1.2.0',
    reason=(
        'This is a low level API. Call class methods of '
        '"pygeonlp.api.dict_manager.DictManager" directly.'
    )
)
def addDictionaryFromWeb(url, params=None, **kwargs):
    """
    指定した URL にあるページに含まれる辞書メタデータ（JSON-LD）を取得し、
    メタデータに記載されている URL から地名解析辞書（CSVファイル）を取得し、
    データベースに登録します。

    既に同じ identifier を持つ辞書データがデータベースに登録されている場合、
    削除してから新しい辞書データを登録します。

    登録した辞書を利用可能にするには、 ``setActivateDictionaries()``
    または ``activateDictionaires()`` で有効化する必要があります。

    Parameters
    ----------
    url : str
        辞書メタデータを含むウェブページの URL。
    params : dict, optional
        requests.get に渡す params パラメータ。
    **kwargs : dict, optional
        requests.get に渡す kwargs パラメータ。

    Returns
    -------
    bool
        常に True。登録に失敗した場合は例外が発生します。
    """  # noqa: E501
    _check_initialized()
    return _default_manager.addDictionaryFromWeb(url, params, **kwargs)


@deprecated(
    version='1.2.0',
    reason=(
        'This is a low level API. Call class methods of '
        '"pygeonlp.api.dict_manager.DictManager" directly.'
    )
)
def saveDictionaryFromWeb(jsonfile, csvfile, url, params=None, **kwargs):
    """
    指定した URL にあるページに含まれる辞書メタデータ（JSON-LD）を取得し、
    メタデータに記載されている URL から地名解析辞書（CSVファイル）を取得し、
    指定されたファイルに保存します。

    Parameters
    ----------
    jsonfile : str
        json-ld を保存するファイル名。
    csvfile : str
        CSV データを保存するファイル名。
    url : str
        辞書メタデータを含むウェブページの URL。
    params : dict, optional
        requests.get に渡す params パラメータ。
    **kwargs : dict, optional
        requests.get に渡す kwargs パラメータ。

    Returns
    -------
    bool
        常に True。ダウンロードに失敗した場合は例外が発生します。

    Examples
    --------
    >>> import pygeonlp.api as api
    >>> api.init()
    >>> api.saveDictionaryFromWeb('geoshape.json', 'geoshape.csv', 'https://geonlp.ex.nii.ac.jp/dictionary/geoshape-city/')
    True
    >>> import os
    >>> os.remove('geoshape.json')
    >>> os.remove('geoshape.csv')
    """  # noqa: E501
    _check_initialized()
    return _default_manager.saveDictionaryFromWeb(jsonfile, csvfile, url,
                                                  params, **kwargs)


@deprecated(
    version='1.2.0',
    reason=(
        'This is a low level API. Call class methods of '
        '"pygeonlp.api.dict_manager.DictManager" directly.'
    )
)
def removeDictionary(identifier):
    """
    identifier で指定した辞書をデータベースから削除します。

    Parameters
    ----------
    identifier : str
        辞書の identifier ("geonlp:xxxxx")。

    Returns
    -------
    bool
        常に True。

    Raises
    ------
    RuntimeError
        指定した辞書が登録されていない場合は RuntimeError が発生します。
    """  # noqa: E501
    _check_initialized()
    return _default_manager.removeDictionary(identifier)


@deprecated(
    version='1.2.0',
    reason=(
        'This is a low level API. Call class methods of '
        '"pygeonlp.api.dict_manager.DictManager" directly.'
    )
)
def updateIndex():
    """
    辞書のインデックスを更新して検索可能にします。
    """
    _check_initialized()
    _default_manager.updateIndex()


@deprecated(
    version='1.2.0',
    reason=(
        'This is a low level API. Call class methods of '
        '"pygeonlp.api.dict_manager.DictManager.setupBasicDatabase" '
        'directly.'
    ))
def setup_basic_database(db_dir=None, src_dir=None):
    """
    基本的な地名解析辞書を登録したデータベースを作成します。

    Parameters
    ----------
    db_dir: PathLike, optional
        データベースディレクトリを指定します。ここにデータベースファイルを作成します。
        既にデータベースが存在する場合は追記します（つまり、既に登録済みの地名解析辞書の
        データは失われません）。

        省略された場合には ``get_db_dir()`` で決定します。
    src_dir: PathLike, optional
        地名解析辞書ファイルが配置されているディレクトリを指定します。
        省略された場合には、 ``sys.prefix`` または ``site.USER_BASE`` の下に
        ``pygeonlp_basedata`` がないか探します。
        見つからない場合は ``RuntimeError`` を送出しますので、
        ディレクトリを指定してください。
    """  # noqa: E501
    manager = DictManager(db_dir=db_dir)
    manager.setupBasicDatabase(src_dir=src_dir)
    del manager


def analyze(sentence, **kwargs):
    """
    文を解析した結果をラティス表現で返します。

    Parameters
    ----------
    sentence : str
        解析するテキスト。

    Returns
    -------
    list
        解析結果のラティス表現。

    Examples
    --------
    >>> import pygeonlp.api as api
    >>> api.init()
    >>> api.analyze('今日は国会議事堂前まで歩きました。')
    [[{"surface": "今日", "node_type": "NORMAL", "morphemes": {"conjugated_form": "*", "conjugation_type": "*", "original_form": "今日", "pos": "名詞", "prononciation": "キョー", "subclass1": "副詞可能", "subclass2": "*", "subclass3": "*", "surface": "今日", "yomi": "キョウ"}, "geometry": null, "prop": null}], [{"surface": "は", "node_type": "NORMAL", "morphemes": {"conjugated_form": "*", "conjugation_type": "*", "original_form": "は", "pos": "助詞", "prononciation": "ワ", "subclass1": "係助詞", "subclass2": "*", "subclass3": "*", "surface": "は", "yomi": "ハ"}, "geometry": null, "prop": null}], [{"surface": "国会議事堂前", "node_type": "GEOWORD", "morphemes": {"conjugated_form": "*", "conjugation_type": "*", "original_form": "国会議事堂前", "pos": "名詞", "prononciation": "", "subclass1": "固有名詞", "subclass2": "地名語", "subclass3": "Bn4q6d:国会議事堂前駅", "surface": "国会議事堂前", "yomi": ""}, "geometry": {"type": "Point", "coordinates": [139.74534166666666, 35.674845]}, "prop": {"body": "国会議事堂前", "dictionary_id": 3, "entry_id": "LrGGxY", "geolod_id": "Bn4q6d", "hypernym": ["東京地下鉄", "4号線丸ノ内線"], "institution_type": "民営鉄道", "latitude": "35.674845", "longitude": "139.74534166666666", "ne_class": "鉄道施設/鉄道駅", "railway_class": "普通鉄道", "suffix": ["駅", ""], "dictionary_identifier": "geonlp:ksj-station-N02"}}, {"surface": "国会議事堂前", "node_type": "GEOWORD", "morphemes": {"conjugated_form": "*", "conjugation_type": "*", "original_form": "国会議事堂前", "pos": "名詞", "prononciation": "", "subclass1": "固有名詞", "subclass2": "地名語", "subclass3": "cE8W4w:国会議事堂前駅", "surface": "国会議事堂前", "yomi": ""}, "geometry": {"type": "Point", "coordinates": [139.74305333333334, 35.673543333333335]}, "prop": {"body": "国会議事堂前", "dictionary_id": 3, "entry_id": "4NFELa", "geolod_id": "cE8W4w", "hypernym": ["東京地下鉄", "9号線千代田線"], "institution_type": "民営鉄道", "latitude": "35.673543333333335", "longitude": "139.74305333333334", "ne_class": "鉄道施設/鉄道駅", "railway_class": "普通鉄道", "suffix": ["駅", ""], "dictionary_identifier": "geonlp:ksj-station-N02"}}], [{"surface": "まで", "node_type": "NORMAL", "morphemes": {"conjugated_form": "*", "conjugation_type": "*", "original_form": "まで", "pos": "助詞", "prononciation": "マデ", "subclass1": "副助詞", "subclass2": "*", "subclass3": "*", "surface": "まで", "yomi": "マデ"}, "geometry": null, "prop": null}], [{"surface": "歩き", "node_type": "NORMAL", "morphemes": {"conjugated_form": "五段・カ行イ音便", "conjugation_type": "連用形", "original_form": "歩く", "pos": "動詞", "prononciation": "アルキ", "subclass1": "自立", "subclass2": "*", "subclass3": "*", "surface": "歩き", "yomi": "アルキ"}, "geometry": null, "prop": null}], [{"surface": "まし", "node_type": "NORMAL", "morphemes": {"conjugated_form": "特殊・マス", "conjugation_type": "連用形", "original_form": "ます", "pos": "助動詞", "prononciation": "マシ", "subclass1": "*", "subclass2": "*", "subclass3": "*", "surface": "まし", "yomi": "マシ"}, "geometry": null, "prop": null}], [{"surface": "た", "node_type": "NORMAL", "morphemes": {"conjugated_form": "特殊・タ", "conjugation_type": "基本形", "original_form": "た", "pos": "助動詞", "prononciation": "タ", "subclass1": "*", "subclass2": "*", "subclass3": "*", "surface": "た", "yomi": "タ"}, "geometry": null, "prop": null}], [{"surface": "。", "node_type": "NORMAL", "morphemes": {"conjugated_form": "*", "conjugation_type": "*", "original_form": "。", "pos": "記号", "prononciation": "。", "subclass1": "句点", "subclass2": "*", "subclass3": "*", "surface": "。", "yomi": "。"}, "geometry": null, "prop": null}]]

    Note
    ----
    ラティス表現では全ての地名語の候補を列挙して返します。
    """  # noqa: E501
    _check_initialized()
    return _default_workflow.parser.analyze(sentence, **kwargs)


def geoparse(sentence):
    """
    Default Workflow を利用して文を解析した結果を、
    GeoJSON Feature 形式に変換可能な dict のリストとして返します。

    Parameters
    ----------
    sentence : str
        解析する文字列。

    Returns
    -------
    list
        形態素で分割した各要素のジオメトリや形態素情報、地名語情報を
        GeoJSON Feature 形式に変換可能な dict で表現し、
        そのリストを返します。

    Examples
    --------
    >>> import json
    >>> import pygeonlp.api as api
    >>> api.init()
    >>> json.dumps({'type':'FeatureCollection', 'features':api.geoparse('今日は国会議事堂前まで歩きました。')}, ensure_ascii=False)
    '{"type": "FeatureCollection", "features": [{"type": "Feature", "geometry": null, "properties": {"surface": "今日", "node_type": "NORMAL", "morphemes": {"conjugated_form": "*", "conjugation_type": "*", "original_form": "今日", "pos": "名詞", "prononciation": "キョー", "subclass1": "副詞可能", "subclass2": "*", "subclass3": "*", "surface": "今日", "yomi": "キョウ"}}}, {"type": "Feature", "geometry": null, "properties": {"surface": "は", "node_type": "NORMAL", "morphemes": {"conjugated_form": "*", "conjugation_type": "*", "original_form": "は", "pos": "助詞", "prononciation": "ワ", "subclass1": "係助詞", "subclass2": "*", "subclass3": "*", "surface": "は", "yomi": "ハ"}}}, {"type": "Feature", "geometry": {"type": "Point", "coordinates": [139.74..., 35.67...]}, "properties": {"surface": "国会議事堂前", "node_type": "GEOWORD", "morphemes": {"conjugated_form": "*", "conjugation_type": "*", "original_form": "国会議事堂前", "pos": "名詞", "prononciation": "", "subclass1": "固有名詞", "subclass2": "地名語", "subclass3": "...:国会議事堂前駅", "surface": "国会議事堂前", "yomi": ""}, "geoword_properties": {"body": "国会議事堂前", "dictionary_id": 3, "entry_id": "...", "geolod_id": "...", "hypernym": ["東京地下鉄", "..."], "institution_type": "民営鉄道", "latitude": "35.67...", "longitude": "139.74...", "ne_class": "鉄道施設/鉄道駅", "railway_class": "普通鉄道", "suffix": ["駅", ""], "dictionary_identifier": "geonlp:ksj-station-..."}}}, {"type": "Feature", "geometry": null, "properties": {"surface": "まで", "node_type": "NORMAL", "morphemes": {"conjugated_form": "*", "conjugation_type": "*", "original_form": "まで", "pos": "助詞", "prononciation": "マデ", "subclass1": "副助詞", "subclass2": "*", "subclass3": "*", "surface": "まで", "yomi": "マデ"}}}, {"type": "Feature", "geometry": null, "properties": {"surface": "歩き", "node_type": "NORMAL", "morphemes": {"conjugated_form": "五段・カ行イ音便", "conjugation_type": "連用形", "original_form": "歩く", "pos": "動詞", "prononciation": "アルキ", "subclass1": "自立", "subclass2": "*", "subclass3": "*", "surface": "歩き", "yomi": "アルキ"}}}, {"type": "Feature", "geometry": null, "properties": {"surface": "まし", "node_type": "NORMAL", "morphemes": {"conjugated_form": "特殊・マス", "conjugation_type": "連用形", "original_form": "ます", "pos": "助動詞", "prononciation": "マシ", "subclass1": "*", "subclass2": "*", "subclass3": "*", "surface": "まし", "yomi": "マシ"}}}, {"type": "Feature", "geometry": null, "properties": {"surface": "た", "node_type": "NORMAL", "morphemes": {"conjugated_form": "特殊・タ", "conjugation_type": "基本形", "original_form": "た", "pos": "助動詞", "prononciation": "タ", "subclass1": "*", "subclass2": "*", "subclass3": "*", "surface": "た", "yomi": "タ"}}}, {"type": "Feature", "geometry": null, "properties": {"surface": "。", "node_type": "NORMAL", "morphemes": {"conjugated_form": "*", "conjugation_type": "*", "original_form": "。", "pos": "記号", "prononciation": "。", "subclass1": "句点", "subclass2": "*", "subclass3": "*", "surface": "。", "yomi": "。"}}}]}'

    Note
    ----
    このメソッドは、文字列の解析、フィルタによる絞り込み、
    ランキングなどの一連の処理を行ないます。
    """  # noqa: E501
    _check_initialized()
    return _default_workflow.geoparse(sentence)


def default_workflow():
    """
    Default Workflow オブジェクトを返します。

    通常はこのメソッドを利用する必要はありません。

    Examples
    --------
    >>> import pygeonlp.api as api
    >>> api.init()
    >>> type(api.default_workflow())
    <class 'pygeonlp.api.workflow.Workflow'>
    """
    return _default_workflow


def _check_initialized():
    """
    Default Workflow オブジェクトが ``init()`` で初期化されていることを確認するための
    プライベートメソッドです。

    この関数は API メソッド内でチェックのために呼びだすものなので、
    ユーザが呼びだす必要は基本的にありません。
    """
    if _default_workflow is None:
        init()
        # raise WorkflowError("APIが初期化されていません。")


def getActiveDictionaries():
    """
    インストール済み辞書のうち、解析に利用する辞書のメタデータ一覧を返します。
    デフォルトでは全てのインストール済み辞書を利用します。

    一時的に辞書を利用したくない場合、 ``disactivateDictionaries()`` で除外できます。
    除外された辞書は ``activateDictionaries()`` で再び利用可能になります。

    Returns
    -------
    list
        Metadata インスタンスのリスト。

    Examples
    --------
    >>> import pygeonlp.api as api
    >>> api.init()
    >>> sorted([x.get_identifier() for x in api.getActiveDictionaries()])
    ['geonlp:geoshape-city', 'geonlp:geoshape-pref', 'geonlp:ksj-station-N02']
    """
    _check_initialized()
    return _default_workflow.getActiveDictionaries()


def setActiveDictionaries(idlist=None, pattern=None):
    """
    インストール済み辞書のうち、解析に利用する辞書を指定します。

    Parameters
    ----------
    idlist : list, optional
        利用する辞書の id または identifier のリスト。
    pattern : str, optional
        利用する辞書の identifier の正規表現。

    Examples
    --------
    >>> import pygeonlp.api as api
    >>> api.init()
    >>> api.setActiveDictionaries(pattern=r'geonlp:geoshape')
    >>> sorted([x.get_identifier() for x in api.getActiveDictionaries()])
    ['geonlp:geoshape-city', 'geonlp:geoshape-pref']

    Note
    ----
    idlist と pattern のどちらかは指定する必要があります。
    """
    _check_initialized()
    return _default_workflow.setActiveDictionaries(idlist, pattern)


def disactivateDictionaries(idlist=None, pattern=None):
    """
    指定した辞書を一時的に除外し、利用しないようにします。
    既に除外されている辞書は除外されたままになります。

    Parameters
    ----------
    idlist : list
        除外する辞書の内部 id または identifier のリスト。
    pattern : str
        除外する辞書の identifier を指定する正規表現。

    Examples
    --------
    >>> import pygeonlp.api as api
    >>> api.init()
    >>> sorted(api.disactivateDictionaries(pattern=r'geonlp:geoshape'))
    ['geonlp:geoshape-city', 'geonlp:geoshape-pref']
    >>> [x.get_identifier() for x in api.getActiveDictionaries()]
    ['geonlp:ksj-station-N02']
    >>> api.disactivateDictionaries(pattern=r'ksj-station')
    ['geonlp:ksj-station-N02']
    >>> [x.get_identifier() for x in api.getActiveDictionaries()]
    []

    Note
    ----
    idlist と pattern を同時に指定した場合、どちらか一方の条件に
    一致する辞書は除外されます。
    """
    _check_initialized()
    return _default_workflow.disactivateDictionaries(idlist, pattern)


def activateDictionaries(idlist=None, pattern=None):
    """
    指定した辞書を再び利用するようにします。
    既に利用可能な辞書は指定しなくても利用可能なままになります。

    Parameters
    ----------
    idlist : list
        利用する辞書の内部 id または identifier のリスト。
    pattern : str
        利用する辞書の identifier を指定する正規表現。

    Examples
    --------
    >>> import pygeonlp.api as api
    >>> api.init()
    >>> sorted(api.disactivateDictionaries(pattern=r'.*'))
    ['geonlp:geoshape-city', 'geonlp:geoshape-pref', 'geonlp:ksj-station-N02']
    >>> sorted([x.get_identifier() for x in api.getActiveDictionaries()])
    []
    >>> api.activateDictionaries(pattern=r'ksj-station')
    ['geonlp:ksj-station-N02']
    >>> sorted([x.get_identifier() for x in api.getActiveDictionaries()])
    ['geonlp:ksj-station-N02']

    Note
    ----
    idlist と pattern を同時に指定した場合、どちらか一方の条件に
    一致する辞書は利用可能になります。
    """
    _check_initialized()
    return _default_workflow.activateDictionaries(idlist, pattern)


def getActiveClasses():
    """
    解析に利用する固有名クラスの一覧を返します。
    デフォルトは '.*' で、全ての固有名クラスが利用されます。

    一時的に特定の固有名クラスだけを解析対象としたい場合、
    setActiveClasses() で対象クラスを指定できます。

    Returns
    -------
    list
        利用する固有名クラス（正規表現）のリスト。

    Examples
    --------
    >>> import pygeonlp.api as api
    >>> api.init()
    >>> api.getActiveClasses()
    ['.*']
    """
    _check_initialized()
    return _default_workflow.getActiveClasses()


def setActiveClasses(patterns=None):
    """
    解析対象とする固有名クラスの正規表現リストを指定します。
    いずれかの正規表現に一致する固有名クラスは解析対象となります。

    '-' から始まる場合、その正規表現に一致する固有名クラスは対象外となります。

    Parameters
    ----------
    patterns : list, optional
        解析対象とする固有名クラス（str）のリスト。
        省略した場合 ['.*'] （全固有名クラス）を対象とします。

    Examples
    --------
    >>> import pygeonlp.api as api
    >>> api.init()
    >>> api.getActiveClasses()
    ['.*']
    >>> api.searchWord('東京都')
    {'QknGsa': {...'dictionary_identifier': 'geonlp:geoshape-pref'}}
    >>> api.setActiveClasses(['.*', '-都道府県'])
    >>> api.searchWord('東京都')
    {}
    >>> api.setActiveClasses()
    >>> api.getActiveClasses()
    ['.*']
    """
    _check_initialized()
    return _default_workflow.setActiveClasses(patterns)


def get_version() -> str:
    """
    Return version string.
    """
    return __version__



from collections.abc import Iterable
import datetime
from logging.config import dictConfig
from typing import List, Union, Optional

from flask import Flask, redirect
from flask_jsonrpc import JSONRPC
from flask_jsonrpc.exceptions import MethodNotFoundError, InvalidParamsError

import jageocoder
import pygeonlp.api as geonlp_api
from pygeonlp.webapi import config

dictConfig(config.LOGGING)
app = Flask('pygeonlp_webapi')
app.config.from_object(config)
jsonrpc = JSONRPC(app, '/api', enable_web_browsable_api=True)


def apply_geonlp_api_parse_options(options: Optional[dict] = None):
    """
    Parse, ParseStructured のオプションに含まれる解析条件を
    geonlp_api に適用します。

    Parameters
    ----------
    options: dict, optional
        Parse オプション

    Notes
    -----
    利用可能なオプションは以下の通りで、この順番に処理されます。

    geocoding: bool
        住所ジオコーディングを利用する場合は true を指定します。

    **辞書の指定**

    set-dic: str or list
        利用する地名解析辞書の identifier のリスト、または正規表現。

    remove-dic: str or list
        除外する地名解析辞書の identifier のリスト、または正規表現。

    add-dic: str or list
        追加する地名解析辞書の identifier のリスト、または正規表現。

    **固有名クラスの指定**

    set-class: list
        解析対象とする固有名クラスの正規表現のリスト。

    add-class: list
        解析対象に追加する固有名クラスの正規表現のリスト。

    remove-class: list
        解析対象から除外する固有名クラスの正規表現のリスト。

    **フィルタの指定**

    geo-contains: str or dict
        GeoContainsFilter の範囲を表す GeoJSON または
        GeoJSON を返す URL。

    geo-disjoint: str or dict
        GeoDisjointFilter を範囲を表す GeoJSON または
        GeoJSON を返す URL。

    time-exists: str, or list of str
        TimeExistsFilter の基準日時を表す文字列、
        または期間の開始日時と終了日時を表す2つの文字列。

    time-before: str, or list of str
        TimeBeforeFilter の基準日時を表す文字列。
        期間が与えられた場合は開始日時を基準日時とします。

    time-after: str, or list of str
        TimeAfterFilter の基準日時を表す文字列。
        期間が与えられた場合は終了日時を基準日時とします。

    time-overlaps: str, or list of str
        TimeOverlapsFilter の基準日時を表す文字列、
        または期間の開始日時と終了日時を表す2つの文字列。

    time-covers: str, or list of str
        TimeCoversFilter の基準日時を表す文字列、
        または期間の開始日時と終了日時を表す2つの文字列。
    """
    if options is None:
        options = {}

    if geonlp_api.default_workflow() is None:
        # 最初のメソッド呼び出しの前に明示的に pygeonlp.api.init() が
        # 呼ばれていない場合、ここで初期化します。
        geonlp_api.init(db_dir=config.GEONLP_DIR,
                        **(config.GEONLP_API_OPTIONS))

    geonlp_api.setActiveDictionaries(pattern=r'.*')  # デフォルトに戻す

    if 'set-dic' in options:
        param = options.get('set-dic')
        if isinstance(param, str):
            geonlp_api.setActiveDictionaries(pattern=param)
        elif isinstance(param, Iterable):
            geonlp_api.setActiveDictionaries(idlist=param)
        else:
            raise TypeError((
                "'set-dic' には利用する辞書の id か identifier のリスト、"
                "または identifer の正規表現文字列を指定してください。"))

    if 'remove-dic' in options:
        param = options.get('remove-dic')
        if isinstance(param, str):
            geonlp_api.disactivateDictionaries(pattern=param)
        elif isinstance(param, Iterable):
            geonlp_api.disactivateDictionaries(idlist=param)
        else:
            raise TypeError((
                "'remove-dic' には除外する辞書の id か identifier のリスト、"
                "または identifer の正規表現文字列を指定してください。"))

    if 'add-dic' in options:
        param = options.get('add-dic')
        if isinstance(param, str):
            geonlp_api.activateDictionaries(pattern=param)
        elif isinstance(param, Iterable):
            geonlp_api.activateDictionaries(idlist=param)
        else:
            raise TypeError((
                "'add-dic' には追加する辞書の id か identifier のリスト、"
                "または identifer の正規表現文字列を指定してください。"))

    active_classes = []
    if 'set-class' in options:
        param = options.get('set-class')
        if isinstance(param, Iterable):
            active_classes = [str(x) for x in list(param)]
        else:
            raise TypeError((
                "'set-class' には解析対象とする固有名クラスを"
                "正規表現のリストで指定してください。"))

    if 'remove-class' in options:
        param = options.get('remove-class')
        if isinstance(param, Iterable):
            for x in list(param):
                if str(x)[0] == '-':
                    active_classes.append(str(x)[1:])
                else:
                    active_classes.append('-' + str(x))

        else:
            raise TypeError((
                "'remove-class' には解析対象から除外する固有名クラスを"
                "正規表現のリストで指定してください。"))

    if 'add-class' in options:
        param = options.get('add-class')
        if isinstance(param, Iterable):
            active_classes += [str(x) for x in list(param)]
        else:
            raise TypeError((
                "'add-class' には解析対象に追加する固有名クラスを"
                "正規表現のリストで指定してください。"))

    geonlp_api.setActiveClasses(active_classes)


def get_filters_from_options(options: Optional[dict] = {}):
    """
    Parse, ParseStructured のオプションに含まれる
    フィルタのリストを解析し、フィルタオブジェクトを作成して返す。
    """
    filters = []

    if 'geo-contains' in options:
        param = options.get('geo-contains')
        if isinstance(param, str):
            geojson = param
        elif isinstance(param, dict) and 'geometry' in param:
            geojson = param['geometry']
        else:
            geojson = param

        from pygeonlp.api.spatial_filter import GeoContainsFilter
        filters.append(GeoContainsFilter(geojson))

    if 'geo-disjoint' in options:
        param = options.get('geo-disjoint')
        if isinstance(param, str):
            geojson = param
        elif isinstance(param, dict) and 'geometry' in param:
            geojson = param['geometry']
        else:
            geojson = param

        from pygeonlp.api.spatial_filter import GeoDisjointFilter
        filters.append(GeoDisjointFilter(geojson))

    if 'time-exists' in options:
        param = options.get('time-exists')
        from pygeonlp.api.temporal_filter import TimeExistsFilter
        if isinstance(param, (str, datetime.date, datetime.datetime)):
            filters.append(TimeExistsFilter(param))
        elif isinstance(param, (list, tuple)):
            filters.append(TimeExistsFilter(*param))
        elif isinstance(param, dict):
            filters.append(TimeExistsFilter(**param))
        else:
            raise TypeError(
                "'time-exists' のパラメータが解析できません")

    if 'time-before' in options:
        param = options.get('time-before')
        from pygeonlp.api.temporal_filter import TimeBeforeFilter
        if isinstance(param, (str, datetime.date, datetime.datetime)):
            filters.append(TimeBeforeFilter(param))
        elif isinstance(param, (list, tuple)):
            filters.append(TimeBeforeFilter(*param))
        elif isinstance(param, dict):
            filters.append(TimeBeforeFilter(**param))
        else:
            raise TypeError(
                "'time-before' のパラメータが解析できません")

    if 'time-after' in options:
        param = options.get('time-after')
        from pygeonlp.api.temporal_filter import TimeAfterFilter
        if isinstance(param, (str, datetime.date, datetime.datetime)):
            filters.append(TimeAfterFilter(param))
        elif isinstance(param, (list, tuple)):
            filters.append(TimeAfterFilter(*param))
        elif isinstance(param, dict):
            filters.append(TimeAfterFilter(**param))
        else:
            raise TypeError(
                "'time-after' のパラメータが解析できません")

    if 'time-overlaps' in options:
        param = options.get('time-overlaps')
        from pygeonlp.api.temporal_filter import TimeOverlapsFilter
        if isinstance(param, (str, datetime.date, datetime.datetime)):
            filters.append(TimeOverlapsFilter(param))
        elif isinstance(param, (list, tuple)):
            filters.append(TimeOverlapsFilter(*param))
        elif isinstance(param, dict):
            filters.append(TimeOverlapsFilter(**param))
        else:
            raise TypeError(
                "'time-overlaps' のパラメータが解析できません")

    if 'time-covers' in options:
        param = options.get('time-covers')
        from pygeonlp.api.temporal_filter import TimeCoversFilter
        if isinstance(param, (str, datetime.date, datetime.datetime)):
            filters.append(TimeCoversFilter(param))
        elif isinstance(param, (list, tuple)):
            filters.append(TimeCoversFilter(*param))
        elif isinstance(param, dict):
            filters.append(TimeCoversFilter(**param))
        else:
            raise TypeError(
                "'time-covers' のパラメータが解析できません")

    return filters


def check_jageocoder_enabled():
    """
    Jageocoder が利用できるかどうか確認します。
    利用できない場合は InvalidParamsError (-32602) を送ります。
    """
    if not config.JAGEOCODER_TREE:
        raise InvalidParamsError(
            message="'geocoding' option is not available on this server.")


@jsonrpc.method('geonlp.version')
def version() -> str:
    """
    pygeonlp のバージョンを返します。

    Returns
    -------
    str
    """
    return geonlp_api.__version__


@jsonrpc.method('geonlp.parse')
def parse(sentence: str, options: Optional[dict] = {}) -> dict:
    """
    テキストを geoparse します。

    Parameters
    ----------
    sentence: str
        変換したいテキスト
    options: dict, optional
        Parse オプション

    Returns
    -------
    dict
        ``features`` に GeoJSON Feature 形式の
        地名語、非地名語、住所をリストとして含む
        dict を返します。
    """
    apply_geonlp_api_parse_options(options)
    filters = get_filters_from_options(options)
    if options.get('geocoding') in (True, 'true', 'True',):
        geocoder = True
    else:
        geocoder = False

    if geocoder:
        check_jageocoder_enabled()

    workflow = geonlp_api.default_workflow()
    workflow.parser.set_jageocoder(geocoder)
    workflow.filters = filters
    result = workflow.geoparse(sentence)

    feature_collection = {
        "type": "FeatureCollection",
        "features": result
    }

    return feature_collection


@jsonrpc.method('geonlp.parseStructured')
def parse_structured(
        sentence_list: List[str],
        options: Optional[dict] = {}) -> dict:
    """
    複数のセンテンスを geoparse します。

    リストのそれぞれのテキストを geoparse してから
    結果を結合します。
    先にテキストを結合して ``parse()`` を呼ぶと、
    テキストが長すぎる場合に自動的に分割するため、
    地名解決の精度が低下する原因になります。

    テキストの意味的な区切り（文、段落など）が
    分かっている場合は、1ブロックずつ ``parse()``
    で処理するか、 ``parseStructured()`` を
    使ってください。

    Parameters
    ----------
    sentence_list: list of str
        変換したいテキストのリスト
    options: dict, optional
        Parse オプション

    Returns
    -------
    dict
        ``features`` に GeoJSON Feature 形式の
        地名語、非地名語、住所をリストとして含む
        dict を返します。
    """
    apply_geonlp_api_parse_options(options)
    filters = get_filters_from_options(options)
    if options.get('geocoding') in (True, 'true', 'True',):
        geocoder = True
    else:
        geocoder = False

    if geocoder:
        check_jageocoder_enabled()

    workflow = geonlp_api.default_workflow()
    workflow.parser.set_jageocoder(geocoder)
    workflow.filters = filters

    result = []
    for sentence in sentence_list:
        result += workflow.geoparse(sentence)

    feature_collection = {
        "type": "FeatureCollection",
        "features": result
    }

    return feature_collection


@jsonrpc.method('geonlp.search')
def search(key: str, options: Optional[dict] = None) -> dict:
    """
    任意の地名語の情報をデータベースから検索します。

    Parameters
    ----------
    key: str
        検索する語の表記または読み
    options: dict
        Parse オプション

    Returns
    -------
    dict
        geolod_id をキー、地名語の情報を値に持つ dict
    """
    apply_geonlp_api_parse_options(options)
    return geonlp_api.searchWord(key)


@jsonrpc.method('geonlp.getGeoInfo')
def get_geo_info(idlist: List[str], options: Optional[dict] = None) -> dict:
    """
    指定した geolod_id を持つ語の情報を返します。
    id がデータベースに存在しない場合は None を返します。

    Parameters
    ----------
    idlist: str or list
        検索する語の geolod_id のリスト
    options: dict
        Parse オプション

    Returns
    -------
    dict
        geolod_id をキー、地名語の情報を値に持つ dict
    """
    apply_geonlp_api_parse_options(options)
    return geonlp_api.getWordInfo(idlist)


@jsonrpc.method('geonlp.getDictionaries')
def get_dictionaries(options: Optional[dict] = None) -> list:
    """
    データベースに登録されている地名解析辞書の
    identifier のリストを返します。

    Parameters
    ----------
    options: dict
        Parse オプション

    Returns
    -------
    list
        辞書 identifier のリスト
    """
    apply_geonlp_api_parse_options(options)
    return sorted(
        [x.get_identifier() for x in geonlp_api.getDictionaries()]
    )


@jsonrpc.method('geonlp.getDictionaryInfo')
def get_dictionary_info(
        identifier: str,
        options: Optional[dict] = None) -> Union[str, None]:
    """
    指定された identifer を持つ地名解析辞書の JSONLD
    メタデータをデータベースから取得します。

    Parameters
    ----------
    identifier: str
        地名解析辞書の identifier
    options: dict
        Parse オプション

    Returns
    -------
    str
        jsonld メタデータ文字列

        identifier と一致する辞書が存在しない場合は None
    """
    apply_geonlp_api_parse_options(options)
    metadata = geonlp_api.getDictionary(identifier)
    if metadata:
        return metadata.jsonld

    return None


@jsonrpc.method('geonlp.addressGeocoding')
def address_geocoding(address: str) -> dict:
    """
    住所ジオコーディング処理を行ないます。

    Parameters
    ----------
    address: str
        住所文字列

    Returns
    -------
    dict
        住所ジオコーディングの結果
    """
    if not config.JAGEOCODER_TREE:
        raise MethodNotFoundError(
            message="'addressGeocoding' is not available on this server.")

    return jageocoder.search(address)


@app.route('/')
def index():
    return redirect('/api/browse')


if __name__ == '__main__':
    app.run(debug=config.DEBUG)
